"""Tool for generating an S matrix automatically from a Tidy3d simulation and terminal port definitions."""

from __future__ import annotations

from typing import Any, Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d import ClipOperation, GeometryGroup, GridSpec, PolySlab
from tidy3d.components.base import cached_property, skip_if_fields_missing
from tidy3d.components.boundary import BroadbandModeABCSpec
from tidy3d.components.frequency_extrapolation import (
    AbstractLowFrequencySmoothingSpec,
    LowFrequencySmoothingSpec,
)
from tidy3d.components.geometry.base import Box
from tidy3d.components.geometry.bound_ops import bounds_union
from tidy3d.components.geometry.utils import _shift_object
from tidy3d.components.geometry.utils_2d import snap_coordinate_to_grid
from tidy3d.components.index import SimulationMap
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.monitor import DirectivityMonitor, ModeMonitor
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.types import Ax, Complex, Coordinate
from tidy3d.components.types.base import annotate_type
from tidy3d.components.viz import add_ax_if_none, equal_aspect
from tidy3d.constants import C_0, MICROMETER, OHM, fp_eps, inf
from tidy3d.exceptions import SetupError, Tidy3dKeyError, ValidationError
from tidy3d.log import log
from tidy3d.plugins.smatrix.component_modelers.base import (
    FWIDTH_FRAC,
    AbstractComponentModeler,
)
from tidy3d.plugins.smatrix.data.data_array import PortDataArray
from tidy3d.plugins.smatrix.ports.base_lumped import AbstractLumpedPort
from tidy3d.plugins.smatrix.ports.coaxial_lumped import CoaxialLumpedPort
from tidy3d.plugins.smatrix.ports.rectangular_lumped import LumpedPort
from tidy3d.plugins.smatrix.ports.types import TerminalPortType
from tidy3d.plugins.smatrix.ports.wave import WavePort
from tidy3d.plugins.smatrix.types import NetworkElement, NetworkIndex, SParamDef

AUTO_RADIATION_MONITOR_NAME = "radiation"
AUTO_RADIATION_MONITOR_BUFFER = 2
AUTO_RADIATION_MONITOR_NUM_POINTS_THETA = 100
AUTO_RADIATION_MONITOR_NUM_POINTS_PHI = 200


class DirectivityMonitorSpec(MicrowaveBaseModel):
    """Specification for automatically generating a :class:`.DirectivityMonitor`.

    Notes
    -----
        When included in the :attr:`.TerminalComponentModeler.radiation_monitors` tuple,
        a :class:`.DirectivityMonitor` will be automatically generated with the specified
        parameters. This allows users to mix manual :class:`.DirectivityMonitor` objects
        with automatically generated ones, each with customizable parameters.

        The default origin (`custom_origin`) for defining observation points in the automatically
        generated monitor is set to (0, 0, 0) in the global coordinate system.

    Example
    -------
    >>> auto_monitor = DirectivityMonitorSpec(
    ...     name="custom_auto",
    ...     buffer=3,
    ...     num_theta_points=50,
    ...     num_phi_points=100
    ... )
    """

    name: Optional[str] = pd.Field(
        None,
        title="Monitor Name",
        description=f"Optional name for the auto-generated monitor. "
        f"If not provided, defaults to '{AUTO_RADIATION_MONITOR_NAME}_' + index of the monitor in the list of radiation monitors.",
    )

    freqs: Optional[tuple[pd.NonNegativeInt, ...]] = pd.Field(
        None,
        title="Frequencies",
        description="Frequencies to obtain fields at. If not provided, uses all frequencies "
        "from the :class:`.TerminalComponentModeler`. Must be a subset of modeler frequencies if provided.",
    )

    buffer: pd.NonNegativeInt = pd.Field(
        AUTO_RADIATION_MONITOR_BUFFER,
        title="Buffer Distance",
        description="Number of grid cells to maintain between monitor and PML/domain boundaries. "
        f"Default: {AUTO_RADIATION_MONITOR_BUFFER} cells.",
    )

    num_theta_points: pd.NonNegativeInt = pd.Field(
        AUTO_RADIATION_MONITOR_NUM_POINTS_THETA,
        title="Elevation Angle Points",
        description="Number of elevation angle (theta) sample points from 0 to π. "
        f"Default: {AUTO_RADIATION_MONITOR_NUM_POINTS_THETA}.",
    )

    num_phi_points: pd.NonNegativeInt = pd.Field(
        AUTO_RADIATION_MONITOR_NUM_POINTS_PHI,
        title="Azimuthal Angle Points",
        description="Number of azimuthal angle (phi) sample points from -π to π. "
        f"Default: {AUTO_RADIATION_MONITOR_NUM_POINTS_PHI}.",
    )

    custom_origin: Optional[Coordinate] = pd.Field(
        (0, 0, 0),
        title="Local Origin",
        description="Local origin used for defining observation points. If ``None``, uses the "
        "monitor's center.",
        units=MICROMETER,
    )


class ModelerLowFrequencySmoothingSpec(AbstractLowFrequencySmoothingSpec):
    """Specifies the low frequency smoothing parameters for the terminal component simulation.
    This specification affects only results at wave ports. Specifically, the mode decomposition data
    for frequencies for which the total simulation time in units of the corresponding period (T = 1/f) is less than
    the specified minimum sampling time will be overridden by extrapolation from the data in the trusted frequency range.
    The trusted frequency range is defined in terms of minimum and maximum sampling times (the total simulation time divided by the corresponding period).

    Example
    -------
    >>> low_freq_smoothing = ModelerLowFrequencySmoothingSpec(
    ...     min_sampling_time=3,
    ...     max_sampling_time=6,
    ...     order=1,
    ...     max_deviation=0.5,
    ... )
    """


class TerminalComponentModeler(AbstractComponentModeler, MicrowaveBaseModel):
    """
    Tool for modeling two-terminal multiport devices and computing port parameters
    with lumped and wave ports.


    Notes
    -----

    **References**

    .. [1]  R. B. Marks and D. F. Williams, "A general waveguide circuit theory,"
            J. Res. Natl. Inst. Stand. Technol., vol. 97, pp. 533, 1992.

    .. [2]  D. M. Pozar, Microwave Engineering, 4th ed. Hoboken, NJ, USA:
            John Wiley & Sons, 2012.
    """

    ports: tuple[TerminalPortType, ...] = pd.Field(
        (),
        title="Terminal Ports",
        description="Collection of lumped and wave ports associated with the network. "
        "For each port, one simulation will be run with a source that is associated with the port.",
    )

    run_only: Optional[tuple[NetworkIndex, ...]] = pd.Field(
        None,
        title="Run Only",
        description="Set of matrix indices that define the simulations to run. "
        "If ``None``, simulations will be run for all indices in the scattering matrix. "
        "If a tuple is given, simulations will be run only for the given matrix indices.",
    )

    element_mappings: tuple[tuple[NetworkElement, NetworkElement, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Tuple of S matrix element mappings, each described by a tuple of "
        "(input_element, output_element, coefficient), where the coefficient is the "
        "element_mapping coefficient describing the relationship between the input and output "
        "matrix element. If all elements of a given column of the scattering matrix are defined "
        "by ``element_mappings``, the simulation corresponding to this column is skipped automatically.",
    )

    radiation_monitors: tuple[
        annotate_type(Union[DirectivityMonitor, DirectivityMonitorSpec]), ...
    ] = pd.Field(
        (),
        title="Radiation Monitors",
        description="Facilitates the calculation of figures-of-merit for antennas. "
        "These monitors will be included in every simulation and record the radiated fields. "
        "Users can specify a combination of :class:`.DirectivityMonitor` objects for manual placement and :class:`.DirectivityMonitorSpec` "
        "objects for automatic generation.",
    )

    assume_ideal_excitation: bool = pd.Field(
        False,
        title="Assume Ideal Excitation",
        description="If ``True``, only the excited port is assumed to have a nonzero incident wave "
        "amplitude power. This choice simplifies the calculation of the scattering matrix. "
        "If ``False``, every entry in the vector of incident wave amplitudes (a) is calculated "
        "explicitly. This choice requires a matrix inversion when calculating the scattering "
        "matrix, but may lead to more accurate scattering parameters when there are "
        "reflections from simulation boundaries. ",
    )

    s_param_def: SParamDef = pd.Field(
        "pseudo",
        title="Scattering Parameter Definition",
        description="Whether to compute scattering parameters using the 'pseudo' or 'power' wave definitions.",
    )

    low_freq_smoothing: Optional[ModelerLowFrequencySmoothingSpec] = pd.Field(
        None,
        title="Low Frequency Smoothing",
        description="The low frequency smoothing parameters for the terminal component simulation.",
    )

    @property
    def _sim_with_sources(self) -> Simulation:
        """Instance of :class:`.Simulation` with all sources and absorbers added for each port, for plotting."""

        sources = [port.to_source(self._source_time) for port in self.ports]
        absorbers = [
            port.to_absorber()
            for port in self.ports
            if isinstance(port, WavePort) and port.absorber
        ]
        return self.simulation.updated_copy(
            sources=sources, internal_absorbers=absorbers, validate=False
        )

    @equal_aspect
    @add_ax_if_none
    def plot_sim(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs: Any,
    ) -> Ax:
        """Plot a :class:`.Simulation` with all sources and absorbers.

        This is a convenience method to visualize the simulation setup for
        troubleshooting. It shows all sources and absorbers for each port.

        Parameters
        ----------
        x : float, optional
            x-coordinate for the cross-section.
        y : float, optional
            y-coordinate for the cross-section.
        z : float, optional
            z-coordinate for the cross-section.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on.
        **kwargs
            Keyword arguments passed to :meth:`.Simulation.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            The axes with the plot.
        """
        return self._sim_with_sources.plot(x=x, y=y, z=z, ax=ax, **kwargs)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs: Any,
    ) -> Ax:
        """Plot permittivity of the :class:`.Simulation`.

        This method shows the permittivity distribution of the simulation with
        all sources and absorbers added for each port.

        Parameters
        ----------
        x : float, optional
            x-coordinate for the cross-section.
        y : float, optional
            y-coordinate for the cross-section.
        z : float, optional
            z-coordinate for the cross-section.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on.
        **kwargs
            Keyword arguments passed to :meth:`.Simulation.plot_eps`.

        Returns
        -------
        matplotlib.axes.Axes
            The axes with the plot.
        """

        return self._sim_with_sources.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    @staticmethod
    def network_index(port: TerminalPortType, mode_index: Optional[int] = None) -> NetworkIndex:
        """Converts the port, and a ``mode_index`` when the port is a :class:`.WavePort`, to a unique string specifier.

        Parameters
        ----------
        port : ``TerminalPortType``
            The port to convert to an index.
        mode_index : Optional[int]
            Selects a single mode from those supported by the ``port``, which is only used when
            the ``port`` is a :class:`.WavePort`

        Returns
        -------
        NetworkIndex
            A unique string that is used to identify the row/column of the scattering matrix.
        """
        return TerminalComponentModeler.get_task_name(port=port, mode_index=mode_index)

    @cached_property
    def network_dict(self) -> dict[NetworkIndex, tuple[TerminalPortType, int]]:
        """Dictionary associating each unique ``NetworkIndex`` to a port and mode index."""
        network_dict = {}
        for port in self.ports:
            if isinstance(port, WavePort):
                for mode_index in port._mode_indices:
                    key = self.network_index(port, mode_index)
                    network_dict[key] = (port, mode_index)
            else:
                key = self.network_index(port, None)
                network_dict[key] = (port, None)
        return network_dict

    @staticmethod
    def _construct_matrix_indices_monitor(
        ports: tuple[TerminalPortType, ...],
    ) -> tuple[NetworkIndex, ...]:
        """Construct matrix indices for monitoring from terminal ports.

        Parameters
        ----------
        ports : tuple[TerminalPortType, ...]
            Tuple of terminal port objects (LumpedPort, CoaxialLumpedPort, or WavePort).

        Returns
        -------
        tuple[NetworkIndex, ...]
            Tuple of network index strings.
        """
        matrix_indices = []
        for port in ports:
            if isinstance(port, WavePort):
                for mode_index in port._mode_indices:
                    matrix_indices.append(TerminalComponentModeler.network_index(port, mode_index))
            else:
                matrix_indices.append(TerminalComponentModeler.network_index(port))
        return tuple(matrix_indices)

    @cached_property
    def matrix_indices_monitor(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the possible matrix indices."""
        return self._construct_matrix_indices_monitor(self.ports)

    @cached_property
    def matrix_indices_source(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the source matrix indices, which may be less than the total number of
        ports."""
        return super().matrix_indices_source

    @cached_property
    def matrix_indices_run_sim(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the matrix indices that will be used to run simulations."""
        return super().matrix_indices_run_sim

    @cached_property
    def sim_dict(self) -> SimulationMap:
        """Generate all the :class:`.Simulation` objects for the port parameter calculation."""

        # Check base simulation for grid size at ports
        TerminalComponentModeler._check_grid_size_at_ports(self.base_sim, self._lumped_ports)
        TerminalComponentModeler._check_grid_size_at_wave_ports(self.base_sim, self._wave_ports)

        sim_dict = {}
        # Now, create simulations with wave port sources and mode solver monitors for computing port modes
        for network_index in self.matrix_indices_run_sim:
            task_name, sim_with_src = self._add_source_to_sim(network_index)
            # update simulation
            sim_dict[task_name] = sim_with_src

        return SimulationMap(keys=tuple(sim_dict.keys()), values=tuple(sim_dict.values()))

    @cached_property
    def _base_sim_no_radiation_monitors(self) -> Simulation:
        """The intermediate base simulation with all grid refinement options, port loads (if present), and monitors added,
        which is only missing the source excitations and radiation monitors.
        """
        # internal mesh override and snapping points are automatically generated from lumped elements.
        lumped_resistors = [port.to_load() for port in self._lumped_ports]

        # Apply the highest frequency in the simulation to define the grid, rather than the
        # source's central frequency, to ensure an accurate solution over the entire range
        grid_spec = self.simulation.grid_spec.copy(
            update={
                "wavelength": C_0 / np.max(self.freqs),
            }
        )

        # Make an initial simulation with new grid_spec to determine where LumpedPorts are snapped
        sim_wo_source = self.simulation.updated_copy(
            grid_spec=grid_spec,
            lumped_elements=lumped_resistors,
            validate=False,
            deep=False,
        )
        snap_centers = {}
        for port in self._lumped_ports:
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                sim_wo_source.grid, port_center_on_axis, port.injection_axis
            )
            snap_centers[port.name] = new_port_center

        # Create monitors and snap to the center positions
        field_monitors = [
            mon
            for port in self.ports
            for mon in port.to_monitors(
                self.freqs, snap_center=snap_centers.get(port.name), grid=sim_wo_source.grid
            )
        ]

        new_mnts = list(self.simulation.monitors) + field_monitors

        new_lumped_elements = list(self.simulation.lumped_elements) + [
            port.to_load(snap_center=snap_centers[port.name]) for port in self._lumped_ports
        ]

        # Add mesh overrides for any wave ports present
        mesh_overrides = list(sim_wo_source.grid_spec.override_structures)
        for wave_port in self._wave_ports:
            if wave_port.num_grid_cells is not None:
                mesh_overrides.extend(wave_port.to_mesh_overrides())
        new_grid_spec = sim_wo_source.grid_spec.updated_copy(override_structures=mesh_overrides)

        new_absorbers = list(sim_wo_source.internal_absorbers)
        for wave_port in self._wave_ports:
            if wave_port.absorber:
                # absorbers are shifted together with sources
                mode_src_pos = wave_port.center[
                    wave_port.injection_axis
                ] + self._shift_value_signed(wave_port)
                port_absorber = wave_port.to_absorber(
                    snap_center=mode_src_pos,
                    freq_spec=BroadbandModeABCSpec(
                        frequency_range=(np.min(self.freqs), np.max(self.freqs))
                    ),
                )
                new_absorbers.append(port_absorber)

        update_dict = {
            "monitors": new_mnts,
            "lumped_elements": new_lumped_elements,
            "grid_spec": new_grid_spec,
            "internal_absorbers": new_absorbers,
        }

        # propagate the low frequency smoothing specification to the simulation
        mode_monitors = [mnt.name for mnt in field_monitors if isinstance(mnt, ModeMonitor)]
        if mode_monitors and self.low_freq_smoothing is not None:
            update_dict["low_freq_smoothing"] = LowFrequencySmoothingSpec(
                monitors=mode_monitors,
                min_sampling_time=self.low_freq_smoothing.min_sampling_time,
                max_sampling_time=self.low_freq_smoothing.max_sampling_time,
                order=self.low_freq_smoothing.order,
                max_deviation=self.low_freq_smoothing.max_deviation,
            )

        # update base simulation with updated set of shared components
        sim_wo_source = sim_wo_source.updated_copy(
            **update_dict,
            validate=False,
            deep=False,
        )

        # extrude port structures
        sim_wo_source = self._extrude_port_structures(sim=sim_wo_source)

        return sim_wo_source

    @cached_property
    def _finalized_radiation_monitors(self) -> tuple[DirectivityMonitor, ...]:
        """
        The tuple of DirectivityMonitor objects for the radiation monitors.

        Expands any DirectivityMonitorSpec instances to actual DirectivityMonitor objects.
        DirectivityMonitor objects are kept as-is.
        """
        base_sim = self._base_sim_no_radiation_monitors
        finalized = []

        for index, rad_mon in enumerate(self.radiation_monitors):
            if isinstance(rad_mon, DirectivityMonitorSpec):
                # Generate DirectivityMonitor from DirectivityMonitorSpec spec
                if not rad_mon.name:
                    mon_name = f"{AUTO_RADIATION_MONITOR_NAME}_{index}"
                    rad_mon = rad_mon.updated_copy(name=mon_name)

                try:
                    generated = self._generate_radiation_monitor(
                        simulation=base_sim, auto_spec=rad_mon
                    )
                    finalized.append(generated)
                except ValueError as e:
                    raise ValueError(
                        "Automatic construction of radiation monitors failed. "
                        "Please address the reason or provide a tuple of DirectivityMonitor "
                        "objects to the 'radiation_monitors' parameter."
                    ) from e
            else:
                # DirectivityMonitor - use as-is
                finalized.append(rad_mon)

        return tuple(finalized)

    @cached_property
    def base_sim(self) -> Simulation:
        """The base simulation with all components added, including radiation monitors."""
        base_sim_tmp = self._base_sim_no_radiation_monitors
        mnts_with_radiation = list(base_sim_tmp.monitors) + list(self._finalized_radiation_monitors)
        grid_spec = GridSpec.from_grid(base_sim_tmp.grid)
        grid_spec.attrs["from_grid_spec"] = base_sim_tmp.grid_spec
        # We skipped validations up to now, here we finally validate the base sim
        return base_sim_tmp.updated_copy(monitors=mnts_with_radiation, grid_spec=grid_spec)

    def _generate_radiation_monitor(
        self, simulation: Simulation, auto_spec: DirectivityMonitorSpec
    ) -> DirectivityMonitor:
        """
        Generates a DirectivityMonitor object for the simulation.

        The monitor is placed at a specified buffer distance from PML boundaries
        (or domain boundaries if no PML). It samples the whole sphere with specified angular resolution.

        The monitor is validated to ensure it is far enough from simulation structures.

        Parameters
        ----------
        simulation : Simulation
            The simulation for which to generate the monitor.
        auto_spec : DirectivityMonitorSpec
            Specification for auto-generation.

        Returns
        -------
        DirectivityMonitor
            The generated monitor configured to measure radiation in all directions.

        Raises
        ------
        ValueError
            If the monitor is not far enough from structures.
        """

        # Extract parameters from auto_spec
        monitor_name = auto_spec.name
        monitor_buffer = auto_spec.buffer
        num_theta = auto_spec.num_theta_points
        num_phi = auto_spec.num_phi_points
        monitor_freqs = auto_spec.freqs or self.freqs

        # Get PML thicknesses in all directions
        pml_layers = simulation.num_pml_layers  # List of (minus, plus) layers for each axis
        grid = simulation.grid
        boundaries = grid.boundaries.to_list  # List of coordinate arrays for each axis
        num_cells = grid.num_cells  # List of number of cells for each axis

        # Calculate monitor span using the specified buffer distance
        mnt_span = [
            (minus_pml + monitor_buffer, num_cells_axis - plus_pml - monitor_buffer)
            for (minus_pml, plus_pml), num_cells_axis in zip(pml_layers, num_cells)
        ]

        # Calculate monitor bounds
        mnt_bounds = [
            (coords[start_idx], coords[end_idx]) if sim_size_axis > 0 else (-inf, inf)
            for (start_idx, end_idx), coords, sim_size_axis in zip(
                mnt_span, boundaries, simulation.size
            )
        ]

        mnt_bounds = np.transpose(mnt_bounds)

        mnt_box = Box.from_bounds(mnt_bounds[0], mnt_bounds[1])

        # Create angle arrays for full sphere sampling
        # theta: elevation angle [0, pi]
        # phi: azimuthal angle [-pi, pi]
        theta = np.linspace(0, np.pi, num_theta)
        phi = np.linspace(-np.pi, np.pi, num_phi)

        # Create the monitor
        monitor = DirectivityMonitor(
            center=mnt_box.center,
            size=mnt_box.size,
            freqs=monitor_freqs,
            name=monitor_name,
            theta=theta,
            phi=phi,
            custom_origin=auto_spec.custom_origin,
        )

        # Validate that monitor is far enough from structures
        self._validate_radiation_monitor_buffer(simulation, mnt_span, monitor_buffer)

        return monitor

    def _validate_radiation_monitor_buffer(
        self, simulation: Simulation, mnt_span: list[tuple[int, int]], buffer: int
    ) -> None:
        """Validate that the radiation monitor is far enough from simulation structures.

        Checks that each side of the monitor is at least AUTO_RADIATION_MONITOR_BUFFER cells
        away from the union of all structures and lumped elements, using grid cell indices.

        Parameters
        ----------
        simulation : Simulation
            The simulation containing structures and lumped elements.
        mnt_span : list[tuple[int, int]]
            The span (start, stop) indices of the monitor in each axis.
        buffer : int
            The buffer distance to use.

        Raises
        ------
        ValueError
            If the monitor is not far enough from structures.
        """
        # Get finalized simulation to include all structures
        finalized_sim = simulation._finalized

        # Get all structures (including finalized ones)
        structures = finalized_sim.structures

        # Get lumped elements
        lumped_elements = simulation.lumped_elements

        # If no structures or lumped elements, validation passes
        if not structures and not lumped_elements:
            return

        # Calculate union of bounding boxes for all structures and lumped elements
        all_geoms = []

        # Add structures
        for struct in structures:
            all_geoms.append(struct.geometry)

        # Add lumped elements (they have geometry)
        for elem in lumped_elements:
            all_geoms.append(elem.to_geometry())

        # Compute union of all bounds
        if all_geoms:
            union_bounds = all_geoms[0].bounds
            for geom in all_geoms[1:]:
                union_bounds = bounds_union(union_bounds, geom.bounds)

            # Convert union bounds to Box and get grid cell indices
            union_box = Box.from_bounds(union_bounds[0], union_bounds[1])
            grid = simulation.grid
            union_inds = grid.discretize_inds(union_box, extend=True)

            # Check each axis
            for axis in range(3):
                mnt_start, mnt_end = mnt_span[axis]
                union_start, union_end = union_inds[axis]

                axis_name = "xyz"[axis]

                # Check minus side: union should be at least BUFFER cells away from monitor start
                buffer_minus = union_start - mnt_start
                if buffer_minus < buffer:
                    raise ValueError(
                        f"Automatically generated radiation monitor is too close to structures on the negative {axis_name} side. "
                        f"Buffer: {buffer_minus} cells, required: {buffer} cells. "
                        f"Please increase simulation domain size."
                    )

                # Check plus side: union should be at least BUFFER cells away from monitor end
                buffer_plus = mnt_end - union_end
                if buffer_plus < buffer:
                    raise ValueError(
                        f"Automatically generated radiation monitor is too close to structures on the positive {axis_name} side. "
                        f"Buffer: {buffer_plus} cells, required: {buffer} cells. "
                        f"Please increase simulation domain size."
                    )

    def _add_source_to_sim(self, source_index: NetworkIndex) -> tuple[str, Simulation]:
        """Adds the source corresponding to the ``source_index`` to the base simulation."""
        port, mode_index = self.network_dict[source_index]
        if isinstance(port, WavePort):
            # Source is placed just before the field monitor of the port
            mode_src_pos = port.center[port.injection_axis] + self._shift_value_signed(port)
            port_source = port.to_source(
                self._source_time, snap_center=mode_src_pos, mode_index=mode_index
            )
        else:
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                self.base_sim.grid, port_center_on_axis, port.injection_axis
            )
            port_source = port.to_source(
                self._source_time, snap_center=new_port_center, grid=self.base_sim.grid
            )
        task_name = self.get_task_name(port=port, mode_index=mode_index)

        return (
            task_name,
            self.base_sim.updated_copy(sources=[port_source], validate=False, deep=False),
        )

    @cached_property
    def _source_time(self):
        """Helper to create a time domain pulse for the frequency range of interest."""
        if self.custom_source_time is not None:
            return self.custom_source_time
        if len(self.freqs) == 1:
            freq0 = self.freqs[0]
            return GaussianPulse(freq0=self.freqs[0], fwidth=freq0 * FWIDTH_FRAC)

        # Using the minimum_source_bandwidth, ensure we don't create a pulse that is too narrowband
        # when fmin and fmax are close together
        return GaussianPulse.from_frequency_range(
            fmin=np.min(self.freqs),
            fmax=np.max(self.freqs),
            remove_dc_component=self.remove_dc_component,
            minimum_source_bandwidth=FWIDTH_FRAC,
        )

    @pd.validator("simulation")
    def _validate_3d_simulation(cls, val):
        """Error if :class:`.Simulation` is not a 3D simulation"""

        if val.size.count(0.0) > 0:
            raise ValidationError(
                f"'{cls.__name__}' must be setup with a 3D simulation with all sizes greater than 0."
            )
        return val

    @pd.validator("ports")
    @skip_if_fields_missing(["simulation"])
    def _validate_port_refinement_usage(cls, val, values):
        """Warn if port refinement options are enabled, but the supplied simulation
        does not contain a grid type that will make use of them."""

        sim: Simulation = values.get("simulation")
        # If grid spec is using AutoGrid
        # then set up is acceptable
        if sim.grid_spec.auto_grid_used:
            return val

        for port in val:
            if port._is_using_mesh_refinement:
                log.warning(
                    f"A port with name '{port.name}' has mesh refinement options enabled, but the "
                    "'Simulation' passed to the 'TerminalComponentModeler' was setup with a 'GridSpec' which "
                    "does not support mesh refinement. For accurate simulations, please setup the "
                    "'Simulation' to use an 'AutoGrid'. To suppress this warning, please explicitly disable "
                    "mesh refinement options in the port, which are by default enabled. For example, set "
                    "the 'enable_snapping_points=False' and 'num_grid_cells=None' for lumped ports."
                )

        return val

    @pd.validator("radiation_monitors")
    @skip_if_fields_missing(["freqs"])
    def _validate_radiation_monitors(cls, val, values):
        """Validate radiation monitors configuration.

        Validates that:
        - DirectivityMonitor frequencies are a subset of modeler frequencies
        - DirectivityMonitorSpec frequencies (if provided) are a subset of modeler frequencies
        """
        modeler_freqs = set(values.get("freqs", []))

        for index, rad_mon in enumerate(val):
            # Only validate freqs if explicitly provided
            # freqs are provided always in DirectivityMonitor
            # in DirectivityMonitorSpec, freqs may be not provided,
            # in this case, we use the modeler frequencies, so no validation is needed
            if rad_mon.freqs is not None:
                mon_freqs = set(rad_mon.freqs)
                is_subset = modeler_freqs.issuperset(mon_freqs)
                if not is_subset:
                    mon_name = rad_mon.name or f"{AUTO_RADIATION_MONITOR_NAME}_{index}"
                    raise ValidationError(
                        f"The frequencies in the radiation monitor '{mon_name}' "
                        f"must be equal to or a subset of the frequencies in the '{cls.__name__}'."
                    )

        return val

    @staticmethod
    def _check_grid_size_at_ports(
        simulation: Simulation, ports: list[Union[LumpedPort, CoaxialLumpedPort]]
    ) -> None:
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        yee_grid = simulation.grid.yee
        for port in ports:
            port._check_grid_size(yee_grid)

    @staticmethod
    def _check_grid_size_at_wave_ports(simulation: Simulation, ports: list[WavePort]) -> None:
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        for port in ports:
            disc_grid = simulation.discretize(port)
            check_axes = port.transverse_axes
            msg_header = f"'WavePort' '{port.name}' "
            for axis in check_axes:
                sim_size = simulation.size[axis]
                dim_cells = disc_grid.num_cells[axis]
                if sim_size > 0 and dim_cells <= 2:
                    small_dim = "xyz"[axis]
                    raise SetupError(
                        msg_header + f"is too small along the "
                        f"'{small_dim}' axis. Less than '3' grid cells were detected. "
                        "Please ensure that the port's 'num_grid_cells' is not 'None'. "
                        "You also may need to use an 'AutoGrid' or `QuasiUniformGrid` "
                        "for the simulation passed to the 'TerminalComponentModeler'."
                    )

    @cached_property
    def _lumped_ports(self) -> list[AbstractLumpedPort]:
        """A list of all lumped ports in the :class:`.TerminalComponentModeler`"""
        return [port for port in self.ports if isinstance(port, AbstractLumpedPort)]

    @cached_property
    def _wave_ports(self) -> list[WavePort]:
        """A list of all wave ports in the :class:`.TerminalComponentModeler`"""
        return [port for port in self.ports if isinstance(port, WavePort)]

    @staticmethod
    def _set_port_data_array_attributes(data_array: PortDataArray) -> PortDataArray:
        """Helper to set additional metadata for ``PortDataArray``."""
        data_array.name = "Z0"
        return data_array.assign_attrs(units=OHM, long_name="characteristic impedance")

    def get_radiation_monitor_by_name(self, monitor_name: str) -> DirectivityMonitor:
        """Find and return a :class:`.DirectivityMonitor` monitor by its name.

        Parameters
        ----------
        monitor_name : str
            Name of the monitor to find.

        Returns
        -------
        :class:`.DirectivityMonitor`
            The monitor matching the given name.

        Raises
        ------
        ``Tidy3dKeyError``
            If no monitor with the given name exists.
        """
        for monitor in self._finalized_radiation_monitors:
            if monitor.name == monitor_name:
                return monitor
        raise Tidy3dKeyError(f"No radiation monitor named '{monitor_name}'.")

    def task_name_from_index(self, source_index: NetworkIndex) -> str:
        """Compute task name for a given network index without constructing simulations."""
        port, mode_index = self.network_dict[source_index]
        return self.get_task_name(port=port, mode_index=mode_index)

    def _extrude_port_structures(self, sim: Simulation) -> Simulation:
        """
        Extrude structures intersecting a port plane when a wave port lies on a structure boundary.

        This method checks wave ports with ``extrude_structures==True`` and automatically extends the boundary structures
        to PEC plates associated with internal absorbers in the direction opposite to the mode source.
        This ensures that mode sources and internal absorbers are fully contained within the extrusion.

        Parameters
        ----------
        sim : Simulation
            Simulation object containing mode sources, internal absorbers, and monitors,
            after mesh overrides and snapping points are applied.

        Returns
        -------
        Simulation
            Updated simulation with extruded structures added to ``simulation.structures``.
        """

        # create list with extruded structures
        new_structures = []
        all_new_structures = []

        # get all mode sources from TerminalComponentModeler that correspond to ports with ``extrude_structures`` flag set to ``True``.
        for port in self.ports:
            if isinstance(port, WavePort) and port.extrude_structures:
                # compute snap_center and shift the internal absorber associated with the current port
                snap_center = port.center[port.injection_axis] + self._shift_value_signed(port)
                absorber = port.to_absorber(snap_center=snap_center)
                shifted_absorber = _shift_object(
                    obj=absorber,
                    grid=sim.grid,
                    bounds=sim.bounds,
                    direction=absorber.direction,
                    shift=absorber.grid_shift,
                )

                # get the PEC box with its face surfaces
                (box, inj_axis, direction) = sim._pec_frame_box(shifted_absorber, expand=True)
                surfaces = box.surfaces(box.size, box.center)

                # get extrusion coordinates and a cutting plane for inference of intersecting structures.
                sign = 1 if direction == "+" else -1
                back_pec_plane = surfaces[2 * inj_axis + (1 if direction == "+" else 0)]

                # get extrusion extent along injection axis
                extrude_to = back_pec_plane.center[inj_axis]

                # move cutting plane beyond the waveport plane along the `ModeSource` injection direction.
                center = list(back_pec_plane.center)
                center[inj_axis] = port.center[inj_axis] - sign * fp_eps * box.size[inj_axis]
                cutting_plane = back_pec_plane.updated_copy(center=center)

                # define extrusion bounds
                extrusion_bounds = [cutting_plane.center[inj_axis], extrude_to][::sign]

                # loop over structures and extrude those that intersect a waveport plane
                for structure in sim.structures:
                    # get geometries that intersect the plane on which the waveport is defined

                    shapely_geom = cutting_plane.intersections_with(structure.geometry)

                    polygon_list = []
                    for geom in shapely_geom:
                        polygon_list = polygon_list + ClipOperation.to_polygon_list(geom)

                    new_geoms = []
                    # loop over identified geometries and extrude them
                    for polygon in polygon_list:
                        # construct outer shell of an extruded geometry first
                        exterior_vertices = np.array(polygon.exterior.coords)
                        outer_shell = PolySlab(
                            axis=inj_axis, slab_bounds=extrusion_bounds, vertices=exterior_vertices
                        )

                        # construct innner shells that represent holes
                        hole_polyslabs = [
                            PolySlab(
                                axis=inj_axis,
                                slab_bounds=extrusion_bounds,
                                vertices=np.array(hole.coords),
                            )
                            for hole in polygon.interiors
                        ]

                        # construct final geometry by removing inner holes from outer shell
                        if hole_polyslabs:
                            holes = GeometryGroup(geometries=hole_polyslabs)
                            extruded_slab_new = ClipOperation(
                                operation="difference", geometry_a=outer_shell, geometry_b=holes
                            )
                        else:
                            extruded_slab_new = outer_shell

                        # append extruded geometry
                        new_geoms.append(extruded_slab_new)
                    if len(polygon_list) != 0:
                        # update structure and add it to the list
                        new_struct = structure.updated_copy(
                            geometry=GeometryGroup(geometries=new_geoms)
                        )
                        new_structures.append(new_struct)

                        # if current port does not intersect any structures raise error
                if not new_structures:
                    raise SetupError(
                        f"The 'WavePort' '{port.name}' does not intersect any structures."
                        f"Please ensure that it is located within or at the boundary of a structure."
                    )

                all_new_structures = all_new_structures + new_structures
                new_structures = []

        # if new structures are extruded (Lumped Port extrusion is ignored)
        if all_new_structures:
            # update structures in simulation while keeping the same grid
            sim = sim.updated_copy(
                grid_spec=GridSpec.from_grid(sim.grid),
                structures=[*sim.structures, *all_new_structures],
                validate=False,
                deep=False,
            )

        return sim


TerminalComponentModeler.update_forward_refs()



"""Utilities for autograd derivative computation and field gradient evaluation."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from functools import reduce
from typing import Any, Callable, Optional

import numpy as np
import xarray as xr

from tidy3d.components.data.data_array import FreqDataArray, ScalarFieldDataArray
from tidy3d.components.data.utils import _zeros_like
from tidy3d.components.types import ArrayLike, Bound, xyz
from tidy3d.config import config
from tidy3d.constants import C_0, EPSILON_0, LARGE_NUMBER, MU_0
from tidy3d.log import log

from .types import PathType
from .utils import get_static

FieldData = dict[str, ScalarFieldDataArray]
PermittivityData = dict[str, ScalarFieldDataArray]
EpsType = FreqDataArray


class LazyInterpolator:
    """Lazy wrapper for interpolators that creates them on first access."""

    def __init__(self, creator_func: Callable) -> None:
        """Initialize with a function that creates the interpolator when called."""
        self.creator_func = creator_func
        self._interpolator = None

    def __call__(self, *args: Any, **kwargs: Any):
        """Create interpolator on first call and delegate to it."""
        if self._interpolator is None:
            self._interpolator = self.creator_func()
        return self._interpolator(*args, **kwargs)


@dataclass
class DerivativeInfo:
    """Stores derivative information passed to the ``._compute_derivatives`` methods.

    This dataclass contains all the field data and parameters needed for computing
    gradients with respect to geometry perturbations.
    """

    # Required fields
    paths: list[PathType]
    """List of paths to the traced fields that need derivatives calculated."""

    E_der_map: FieldData
    """Electric field gradient map.
    Dataset where the field components ("Ex", "Ey", "Ez") store the multiplication
    of the forward and adjoint electric fields. The tangential components of this
    dataset are used when computing adjoint gradients for shifting boundaries.
    All components are used when computing volume-based gradients."""

    D_der_map: FieldData
    """Displacement field gradient map.
    Dataset where the field components ("Ex", "Ey", "Ez") store the multiplication
    of the forward and adjoint displacement fields. The normal component of this
    dataset is used when computing adjoint gradients for shifting boundaries."""

    E_fwd: FieldData
    """Forward electric fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the forward
    electric fields used for computing gradients for a given structure."""

    E_adj: FieldData
    """Adjoint electric fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the adjoint
    electric fields used for computing gradients for a given structure."""

    D_fwd: FieldData
    """Forward displacement fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the forward
    displacement fields used for computing gradients for a given structure."""

    D_adj: FieldData
    """Adjoint displacement fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the adjoint
    displacement fields used for computing gradients for a given structure."""

    eps_data: PermittivityData
    """Permittivity dataset.
    Dataset of relative permittivity values along all three dimensions.
    Used for automatically computing permittivity inside or outside of a simple geometry."""

    eps_in: EpsType
    """Permittivity inside the Structure.
    Typically computed from Structure.medium.eps_model.
    Used when it cannot be computed from eps_data or when eps_approx=True."""

    eps_out: EpsType
    """Permittivity outside the Structure.
    Typically computed from Simulation.medium.eps_model.
    Used when it cannot be computed from eps_data or when eps_approx=True."""

    bounds: Bound
    """Geometry bounds.
    Bounds corresponding to the structure, used in Medium calculations."""

    bounds_intersect: Bound
    """Geometry and simulation intersection bounds.
    Bounds corresponding to the minimum intersection between the structure
    and the simulation it is contained in."""

    simulation_bounds: Bound
    """Simulation bounds.
    Bounds corresponding to the simulation domain containing this structure.
    Unlike bounds_intersect, this is independent of the structure's bounds and
    is purely based on the simulation geometry."""

    frequencies: ArrayLike
    """Frequencies at which the adjoint gradient should be computed."""

    # Optional fields with defaults

    H_der_map: Optional[FieldData] = None
    """Magnetic field gradient map.
    Dataset where the field components ("Hx", "Hy", "Hz") store the multiplication
    of the forward and adjoint magnetic fields. The tangential component of this
    dataset is used when computing adjoint gradients for shifting boundaries of
    structures composed of PEC mediums."""

    H_fwd: Optional[FieldData] = None
    """Forward magnetic fields.
    Dataset where the field components ("Hx", "Hy", "Hz") represent the forward
    magnetic fields used for computing gradients for a given structure."""

    H_adj: Optional[FieldData] = None
    """Adjoint magnetic fields.
    Dataset where the field components ("Hx", "Hy", "Hz") represent the adjoint
    magnetic fields used for computing gradients for a given structure."""

    is_medium_pec: bool = False
    """Indicates if structure material is PEC.
    If True, the structure contains a PEC material which changes the gradient
    formulation at the boundary compared to the dielectric case."""

    background_medium_is_pec: bool = False
    """Indicates if structure material is PEC.
    If True, the structure is partially surrounded by a PEC material."""

    interpolators: Optional[dict] = None
    """Pre-computed interpolators.
    Optional pre-computed interpolators for field components and permittivity data.
    When provided, avoids redundant interpolator creation for multiple geometries
    sharing the same field data. This significantly improves performance for
    GeometryGroup processing."""

    # private cache for interpolators
    _interpolators_cache: dict = field(default_factory=dict, init=False, repr=False)

    def updated_copy(self, **kwargs: Any):
        """Create a copy with updated fields."""
        kwargs.pop("deep", None)
        kwargs.pop("validate", None)
        return replace(self, **kwargs)

    @staticmethod
    def _nan_to_num_if_needed(coords: np.ndarray) -> np.ndarray:
        """Convert NaN and infinite values to finite numbers, optimized for finite inputs."""
        # skip check for small arrays
        if coords.size < 1000:
            return np.nan_to_num(coords, posinf=LARGE_NUMBER, neginf=-LARGE_NUMBER)

        if np.isfinite(coords).all():
            return coords
        return np.nan_to_num(coords, posinf=LARGE_NUMBER, neginf=-LARGE_NUMBER)

    @staticmethod
    def _evaluate_with_interpolators(
        interpolators: dict, coords: np.ndarray
    ) -> dict[str, np.ndarray]:
        """Evaluate field components at coordinates using cached interpolators.

        Parameters
        ----------
        interpolators : dict
            Dictionary mapping field component names to ``RegularGridInterpolator`` objects.
        coords : np.ndarray
            Spatial coordinates (N, 3) where fields are evaluated.

        Returns
        -------
        dict[str, np.ndarray]
            Dictionary mapping component names to field values at coordinates.
        """
        auto_cfg = config.adjoint
        float_dtype = auto_cfg.gradient_dtype_float
        complex_dtype = auto_cfg.gradient_dtype_complex

        coords = DerivativeInfo._nan_to_num_if_needed(coords)
        if coords.dtype != float_dtype and coords.dtype != complex_dtype:
            coords = coords.astype(float_dtype, copy=False)
        return {name: interp(coords) for name, interp in interpolators.items()}

    def create_interpolators(self, dtype: Optional[np.dtype] = None) -> dict:
        """Create interpolators for field components and permittivity data.

        Creates and caches ``RegularGridInterpolator`` objects for all field components
        (E_fwd, E_adj, D_fwd, D_adj) and permittivity data (eps_inf, eps_no).
        This caching strategy significantly improves performance by avoiding
        repeated interpolator construction in gradient evaluation loops.

        Parameters
        ----------
        dtype : np.dtype, optional
            Data type for interpolation coordinates and values. Defaults to the
            current ``config.adjoint.gradient_dtype_float``.

        Returns
        -------
        dict
            Nested dictionary structure:
            - Field data: {"E_fwd": {"Ex": interpolator, ...}, ...}
            - Permittivity: {"eps_inf": interpolator, "eps_no": interpolator}
        """
        from scipy.interpolate import RegularGridInterpolator

        auto_cfg = config.adjoint
        if dtype is None:
            dtype = auto_cfg.gradient_dtype_float
        complex_dtype = auto_cfg.gradient_dtype_complex

        cache_key = str(dtype)
        if cache_key in self._interpolators_cache:
            return self._interpolators_cache[cache_key]

        interpolators = {}
        coord_cache = {}

        def _make_lazy_interpolator_group(
            field_data_dict, group_key, is_field_group=True, override_method: Optional[str] = None
        ) -> None:
            """Helper to create a group of lazy interpolators."""
            if is_field_group:
                interpolators[group_key] = {}

            for component_name, arr in field_data_dict.items():
                # use object ID for caching to handle shared grids
                arr_id = id(arr.data)
                if arr_id not in coord_cache:
                    points = tuple(c.data.astype(dtype, copy=False) for c in (arr.x, arr.y, arr.z))
                    coord_cache[arr_id] = points
                points = coord_cache[arr_id]

                def creator_func(arr=arr, points=points):
                    data = arr.data.astype(
                        complex_dtype if np.iscomplexobj(arr.data) else dtype, copy=False
                    )
                    # create interpolator with frequency dimension
                    if "f" in arr.dims:
                        freq_coords = arr.coords["f"].data.astype(dtype, copy=False)
                        # ensure frequency dimension is last
                        if arr.dims != ("x", "y", "z", "f"):
                            freq_dim_idx = arr.dims.index("f")
                            axes = list(range(data.ndim))
                            axes.append(axes.pop(freq_dim_idx))
                            data = np.transpose(data, axes)
                    else:
                        # single frequency case - add singleton dimension
                        freq_coords = np.array([0.0], dtype=dtype)
                        data = data[..., np.newaxis]

                    points_with_freq = (*points, freq_coords)
                    # If PEC, use nearest interpolation instead of linear to avoid interpolating
                    # with field values inside the PEC (which are 0). Instead, we make sure to
                    # choose interpolation points such that their nearest location is outside of
                    # the PEC surface. The same applies if the background_medium is marked as PEC
                    # since we will need to use the same interpolation strategy inside the structure
                    # border.
                    method = (
                        "nearest"
                        if (self.is_medium_pec or self.background_medium_is_pec)
                        else "linear"
                    )
                    if override_method is not None:
                        method = override_method
                    interpolator_obj = RegularGridInterpolator(
                        points_with_freq, data, method=method, bounds_error=False, fill_value=None
                    )

                    def interpolator(coords):
                        # coords: (N, 3) spatial points
                        n_points = coords.shape[0]
                        n_freqs = len(freq_coords)

                        # build coordinates with frequency dimension
                        coords_with_freq = np.empty((n_points * n_freqs, 4), dtype=coords.dtype)
                        coords_with_freq[:, :3] = np.repeat(coords, n_freqs, axis=0)
                        coords_with_freq[:, 3] = np.tile(freq_coords, n_points)

                        result = interpolator_obj(coords_with_freq)
                        return result.reshape(n_points, n_freqs)

                    return interpolator

                if is_field_group:
                    interpolators[group_key][component_name] = LazyInterpolator(creator_func)
                else:
                    interpolators[component_name] = LazyInterpolator(creator_func)

        # process field interpolators (nested dictionaries)
        interpolator_groups = [
            ("E_fwd", self.E_fwd),
            ("E_adj", self.E_adj),
            ("D_fwd", self.D_fwd),
            ("D_adj", self.D_adj),
        ]
        if self.is_medium_pec or self.background_medium_is_pec:
            interpolator_groups += [("H_fwd", self.H_fwd), ("H_adj", self.H_adj)]
        for group_key, data_dict in interpolator_groups:
            _make_lazy_interpolator_group(
                data_dict, f"{group_key}_linear", is_field_group=True, override_method="linear"
            )
            _make_lazy_interpolator_group(
                data_dict, f"{group_key}_nearest", is_field_group=True, override_method="nearest"
            )

        if self.eps_data is not None:
            _make_lazy_interpolator_group(
                self.eps_data, "eps_data", is_field_group=True, override_method="nearest"
            )

        if self.eps_in is not None:
            _make_lazy_interpolator_group(
                {"eps_in": self.eps_in}, None, is_field_group=False, override_method="nearest"
            )
        if self.eps_out is not None:
            _make_lazy_interpolator_group(
                {"eps_out": self.eps_out}, None, is_field_group=False, override_method="nearest"
            )

        self._interpolators_cache[cache_key] = interpolators
        return interpolators

    def evaluate_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: Optional[dict] = None,
    ) -> np.ndarray:
        """Compute adjoint gradients at surface points for shape optimization.

        Implements the surface integral formulation for computing gradients with respect
        to geometry perturbations.

        Parameters
        ----------
        spatial_coords : np.ndarray
            (N, 3) array of surface evaluation points.
        normals : np.ndarray
            (N, 3) array of outward-pointing normal vectors at each surface point.
        perps1 : np.ndarray
            (N, 3) array of first tangent vectors perpendicular to normals.
        perps2 : np.ndarray
            (N, 3) array of second tangent vectors perpendicular to both normals and perps1.
        interpolators : dict = None
            Pre-computed field interpolators for efficiency.

        Returns
        -------
        np.ndarray
            (N,) array of gradient values at each surface point. Must be integrated
            with appropriate quadrature weights to get total gradient.
        """
        if interpolators is None:
            raise NotImplementedError(
                "Direct field evaluation without interpolators is not implemented. "
                "Please create interpolators using 'create_interpolators()' first."
            )

        # In all paths below, we need to have computed the gradient integration for a
        # dielectric-dielectric interface.
        vjps_dielectric = self._evaluate_dielectric_gradient_at_points(
            spatial_coords,
            normals,
            perps1,
            perps2,
            interpolators,
            self.eps_in,
            self.eps_out,
        )

        if self.is_medium_pec:
            # The structure medium is PEC, but there may be a part of the interface that has
            # dielectric placed on top of or around it where we want to use the dielectric
            # gradient integration. We use the mask to choose between the PEC-dielectric and
            # dielectric-dielectric parts of the border.

            # Detect PEC by looking just inside the boundary
            mask_pec = self._detect_pec_gradient_points(
                spatial_coords,
                normals,
                self.eps_in,
                interpolators["eps_data"],
                is_outside=False,
            )

            # Compute PEC gradients, pulling fields outside of the boundary
            vjps_pec = self._evaluate_pec_gradient_at_points(
                spatial_coords,
                normals,
                perps1,
                perps2,
                interpolators,
                ("eps_out", self.eps_out),
                is_outside=True,
            )

            vjps = mask_pec * vjps_pec + (1.0 - mask_pec) * vjps_dielectric
        elif self.background_medium_is_pec:
            # The structure medium is dielectric, but there may be a part of the interface that has
            # PEC placed on top of or around it where we want to use the PEC gradient integration.
            # We use the mask to choose between the dielectric-dielectric and PEC-dielectric parts
            # of the border.

            # Detect PEC by looking just outside the boundary
            mask_pec = self._detect_pec_gradient_points(
                spatial_coords,
                normals,
                self.eps_out,
                interpolators["eps_data"],
                is_outside=True,
            )

            # Compute PEC gradients, pulling fields inside of the boundary and applying a negative
            # sign compared to above because inside and outside definitions are switched
            vjps_pec = -self._evaluate_pec_gradient_at_points(
                spatial_coords,
                normals,
                perps1,
                perps2,
                interpolators,
                ("eps_in", self.eps_in),
                is_outside=False,
            )

            vjps = mask_pec * vjps_pec + (1.0 - mask_pec) * vjps_dielectric
        else:
            # The structure and its background are both assumed to be dielectric, so we use the
            # dielectric-dielectric gradient integration.
            vjps = vjps_dielectric

        # sum over frequency dimension
        vjps = np.sum(vjps, axis=-1)

        return vjps

    def _evaluate_dielectric_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: dict,
        eps_in_data: ScalarFieldDataArray,
        eps_out_data: ScalarFieldDataArray,
    ) -> np.ndarray:
        eps_out_coords = self._snap_spatial_coords_boundary(
            spatial_coords,
            normals,
            is_outside=True,
            data_array=eps_out_data,
        )
        eps_in_coords = self._snap_spatial_coords_boundary(
            spatial_coords,
            normals,
            is_outside=False,
            data_array=eps_in_data,
        )

        eps_out = interpolators["eps_out"](eps_out_coords)
        eps_in = interpolators["eps_in"](eps_in_coords)

        # evaluate all field components at surface points
        E_fwd_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["E_fwd_linear"].items()
        }
        E_adj_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["E_adj_linear"].items()
        }
        D_fwd_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["D_fwd_linear"].items()
        }
        D_adj_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["D_adj_linear"].items()
        }

        delta_eps_inv = 1.0 / eps_in - 1.0 / eps_out
        delta_eps = eps_in - eps_out

        # project fields onto local surface basis (normal + two tangents)
        D_fwd_norm = self._project_in_basis(D_fwd_at_coords, basis_vector=normals)
        D_adj_norm = self._project_in_basis(D_adj_at_coords, basis_vector=normals)

        E_fwd_perp1 = self._project_in_basis(E_fwd_at_coords, basis_vector=perps1)
        E_adj_perp1 = self._project_in_basis(E_adj_at_coords, basis_vector=perps1)

        E_fwd_perp2 = self._project_in_basis(E_fwd_at_coords, basis_vector=perps2)
        E_adj_perp2 = self._project_in_basis(E_adj_at_coords, basis_vector=perps2)

        D_der_norm = D_fwd_norm * D_adj_norm
        E_der_perp1 = E_fwd_perp1 * E_adj_perp1
        E_der_perp2 = E_fwd_perp2 * E_adj_perp2

        vjps = -delta_eps_inv * D_der_norm + E_der_perp1 * delta_eps + E_der_perp2 * delta_eps

        return vjps

    def _snap_spatial_coords_boundary(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        is_outside: bool,
        data_array: ScalarFieldDataArray,
    ) -> np.ndarray:
        """Assuming a nearest interpolation, adjust the interpolation points given the grid
        defined by `grid_centers` and using `spatial_coords` as a starting point such that we
        select a point inside/outside the boundary depending on is_outside.

             *** (nearest point outside boundary)
              ^
              | n (normal direction)
              |
        _.-~'`-._.-~'`-._ (boundary)
              * (nearest point)

        Parameters
        ----------
        spatial_coords : np.ndarray
            (N, 3) array of surface evaluation points.
        normals : np.ndarray
            (N, 3) array of outward-pointing normal vectors at each surface point.
        is_outside: bool
            Indicator specifying if coordinates should be snapped inside or outside the boundary.
        data_array: ScalarFieldDataArray
            Data array to pull grid centers from when snapping coordinates.

        Returns
        -------
        np.ndarray
            (N, 3) array of coordinate centers at which to interpolate such that they line up
            with a grid center and are inside/outside the boundary
        """
        coords = data_array.coords
        grid_centers = {key: np.array(coords[key].values) for key in coords}

        grid_ddim = np.zeros_like(normals)
        for idx, dim in enumerate("xyz"):
            expanded_coords = np.expand_dims(spatial_coords[:, idx], axis=1)
            grid_centers_select = grid_centers[dim]

            diff = np.abs(expanded_coords - grid_centers_select)

            nearest_grid = np.argmin(diff, axis=-1)
            nearest_grid = np.minimum(np.maximum(nearest_grid, 1), len(grid_centers_select) - 1)

            # compute the local grid spacing near the boundary
            grid_ddim[:, idx] = (
                grid_centers_select[nearest_grid] - grid_centers_select[nearest_grid - 1]
            )

        #
        # Assuming we move in the normal direction, finds which dimension we need to move the least
        # in order to ensure we snap to a point outside the boundary in the worst case (i.e. - the
        # nearest point is just inside the surface)
        #
        # Cover for 2D cases using filter below:
        # 2D case 1:
        #    - in plane gradients where normal: [a, b, 0] and grid: [dx, dy, 0]
        #    - want to rely on in plane normals for boundary snapping (filter on normal component = 0)
        # 2D case 2:
        #    - out of plane gradietns where normal: [0, 0, 1] and grid: [dx, dy, 0]
        #    - want to rely on out of plane normal (so do not want to filter on grid component = 0)
        #    - data may not be captured out of plane, so no snapping will occur even with coords_dn = 0
        #
        small_number = np.finfo(normals.dtype).eps
        coords_dn = np.min(
            np.where(
                (np.abs(normals) > small_number),
                np.abs(grid_ddim) / (np.abs(normals) + small_number),
                np.inf,
            ),
            axis=1,
            keepdims=True,
        )

        # adjust coordinates by half a grid point outside boundary such that nearest interpolation
        # point snaps to outside the boundary
        normal_direction = 1.0 if is_outside else -1.0
        adjust_spatial_coords = (
            spatial_coords
            + normal_direction * normals * config.adjoint.boundary_snapping_fraction * coords_dn
        )

        return adjust_spatial_coords

    def _compute_edge_distance(
        self,
        spatial_coords: np.ndarray,
        grid_centers: dict[str, np.ndarray],
        adjust_spatial_coords: np.ndarray,
    ) -> np.ndarray:
        """Assuming nearest neighbor interpolation, computes the edge distance after interpolation when using the
        adjust_spatial_coords computed from _snap_spatial_coords_boundary.

        Parameters
        ----------
        spatial_coords : np.ndarray
            (N, 3) array of surface evaluation points.
        normals : np.ndarray
            (N, 3) array of outward-pointing normal vectors at each surface point.
        grid_centers: dict[str, np.ndarray]
            The grid points for a given field component indexed by dimension. These grid points
            are used to find the nearest snapping point and adjust the interpolation coordinates
            to ensure we fall inside/outside of a boundary.

        Returns
        -------
        np.ndarray
            (N,) array of distances from the nearest interpolation points to the desired surface
            edge points specified by `spatial_coords`
        """

        edge_distance_squared_sum = np.zeros_like(adjust_spatial_coords[:, 0])
        for idx, dim in enumerate("xyz"):
            expanded_adjusted_coords = np.expand_dims(adjust_spatial_coords[:, idx], axis=1)
            grid_centers_select = grid_centers[dim]

            # find nearest grid point from the adjusted coordinates
            diff = np.abs(expanded_adjusted_coords - grid_centers_select)
            nearest_grid = np.argmin(diff, axis=-1)

            # compute edge distance from the nearest interpolated point to the boundary edge
            edge_distance_squared_sum += (
                np.abs(spatial_coords[:, idx] - grid_centers_select[nearest_grid]) ** 2
            )

        # this edge distance is useful when correcting for edge singularities like those from a PEC
        # material and is used when the PEC PolySlab structure has zero thickness, for example
        edge_distance = np.sqrt(edge_distance_squared_sum)

        return edge_distance

    def _detect_pec_gradient_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        eps_data: ScalarFieldDataArray,
        interpolator: LazyInterpolator,
        is_outside: bool,
    ):
        def _detect_pec(eps_mask):
            return 1.0 * (eps_mask < config.adjoint.pec_detection_threshold)

        adjusted_coords = self._snap_spatial_coords_boundary(
            spatial_coords=spatial_coords,
            normals=normals,
            is_outside=is_outside,
            data_array=eps_data,
        )

        eps_adjusted_all = [
            component_interpolator(adjusted_coords)
            for _, component_interpolator in interpolator.items()
        ]
        eps_detect_pec = reduce(np.minimum, eps_adjusted_all)

        return _detect_pec(eps_detect_pec)

    def _evaluate_pec_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: dict,
        eps_dielectric: tuple[str, ScalarFieldDataArray],
        is_outside: bool,
    ) -> np.ndarray:
        eps_dielectric_key, eps_dielectric_data = eps_dielectric

        def _snap_coordinate_outside(field_components: FieldData):
            """Helper function to perform coordinate adjustment and compute edge distance for each
            component in `field_components`.

            Parameters
            ----------
            field_components: FieldData
                The field components (i.e - Ex, Ey, Ez, Hx, Hy, Hz) that we would like to sample just
                outside the PEC surface using nearest interpolation.

            Returns
            -------
            dict[str, dict[str, np.ndarray]]
                Dictionary mapping each field component name to a dictionary of adjusted coordinates
                and edge distances for that component.
            """
            adjustment = {}
            for name in field_components:
                field_component = field_components[name]
                field_component_coords = field_component.coords

                grid_centers = {
                    key: np.array(field_component_coords[key].values)
                    for key in field_component_coords
                }

                adjusted_coords = self._snap_spatial_coords_boundary(
                    spatial_coords,
                    normals,
                    is_outside=is_outside,
                    data_array=field_component,
                )

                edge_distance = self._compute_edge_distance(
                    spatial_coords=spatial_coords,
                    grid_centers=grid_centers,
                    adjust_spatial_coords=adjusted_coords,
                )
                adjustment[name] = {"coords": adjusted_coords, "edge_distance": edge_distance}

            return adjustment

        def _interpolate_field_components(interp_coords, field_name):
            return {
                name: interp(interp_coords[name]["coords"])
                for name, interp in interpolators[field_name].items()
            }

        # adjust coordinates for PEC to be outside structure bounds and get edge distance for singularity correction.
        E_fwd_coords_adjusted = _snap_coordinate_outside(self.E_fwd)
        E_adj_coords_adjusted = _snap_coordinate_outside(self.E_adj)

        H_fwd_coords_adjusted = _snap_coordinate_outside(self.H_fwd)
        H_adj_coords_adjusted = _snap_coordinate_outside(self.H_adj)

        # using the adjusted coordinates, evaluate all field components at surface points
        E_fwd_at_coords = _interpolate_field_components(
            E_fwd_coords_adjusted, field_name="E_fwd_nearest"
        )
        E_adj_at_coords = _interpolate_field_components(
            E_adj_coords_adjusted, field_name="E_adj_nearest"
        )
        H_fwd_at_coords = _interpolate_field_components(
            H_fwd_coords_adjusted, field_name="H_fwd_nearest"
        )
        H_adj_at_coords = _interpolate_field_components(
            H_adj_coords_adjusted, field_name="H_adj_nearest"
        )

        eps_coords_adjusted = self._snap_spatial_coords_boundary(
            spatial_coords,
            normals,
            is_outside=is_outside,
            data_array=eps_dielectric_data,
        )
        eps_dielectric = interpolators[eps_dielectric_key](eps_coords_adjusted)

        structure_sizes = np.array(
            [self.bounds[1][idx] - self.bounds[0][idx] for idx in range(len(self.bounds[0]))]
        )

        is_flat_perp_dim1 = np.isclose(np.abs(np.sum(perps1[0] * structure_sizes)), 0.0)
        is_flat_perp_dim2 = np.isclose(np.abs(np.sum(perps2[0] * structure_sizes)), 0.0)
        flat_perp_dims = [is_flat_perp_dim1, is_flat_perp_dim2]

        # check if this integration is happening along an edge in which case we will eliminate
        # on of the H field integration components and apply singularity correction
        pec_line_integration = is_flat_perp_dim1 or is_flat_perp_dim2

        def _compute_singularity_correction(adjustment_: dict[str, dict[str, np.ndarray]]):
            """
            Given the `adjustment_` which contains the distance from the PEC edge each field
            component is nearest interpolated at, computes the singularity correction when
            working with 2D PEC using the average edge_distance for each component. In the case
            of 3D PEC gradients, no singularity correction is applied so an array of ones is returned.

            Parameters
            ----------
            adjustment_: dict[str, dict[str, np.ndarray]]
                Dictionary that maps field component name to a dictionary containing the coordinate
                adjustment and the distance to the PEC edge for those coordinates. The edge distance
                is used for 2D PEC singularity correction.

            Returns
            -------
            np.ndarray
                Returns the singularity correction which has shape (N,) where there are N points in
                `spatial_coords`
            """
            return (
                (
                    0.5
                    * np.pi
                    * np.mean([adjustment_[name]["edge_distance"] for name in adjustment_], axis=0)
                )
                if pec_line_integration
                else np.ones_like(spatial_coords, shape=spatial_coords.shape[0])
            )

        E_norm_singularity_correction = np.expand_dims(
            _compute_singularity_correction(E_fwd_coords_adjusted), axis=1
        )
        H_perp_singularity_correction = np.expand_dims(
            _compute_singularity_correction(H_fwd_coords_adjusted), axis=1
        )

        E_fwd_norm = self._project_in_basis(E_fwd_at_coords, basis_vector=normals)
        E_adj_norm = self._project_in_basis(E_adj_at_coords, basis_vector=normals)

        # compute the normal E contribution to the gradient (the tangential E contribution
        # is 0 in the case of PEC since this field component is continuous and thus 0 at
        # the boundary)
        contrib_E = E_norm_singularity_correction * eps_dielectric * E_fwd_norm * E_adj_norm
        vjps = contrib_E

        # compute the tangential H contribution to the gradient (the normal H contribution
        # is 0 for PEC)
        H_fwd_perp1 = self._project_in_basis(H_fwd_at_coords, basis_vector=perps1)
        H_adj_perp1 = self._project_in_basis(H_adj_at_coords, basis_vector=perps1)

        H_fwd_perp2 = self._project_in_basis(H_fwd_at_coords, basis_vector=perps2)
        H_adj_perp2 = self._project_in_basis(H_adj_at_coords, basis_vector=perps2)

        H_der_perp1 = H_perp_singularity_correction * H_fwd_perp1 * H_adj_perp1
        H_der_perp2 = H_perp_singularity_correction * H_fwd_perp2 * H_adj_perp2

        H_integration_components = (H_der_perp1, H_der_perp2)
        if pec_line_integration:
            # if we are integrating along the line, we choose the H component normal to
            # the edge which corresponds to a surface current along the edge whereas the other
            # tangential component corresponds to a surface current along the flat dimension.
            H_integration_components = tuple(
                H_comp for idx, H_comp in enumerate(H_integration_components) if flat_perp_dims[idx]
            )

        # for each of the tangential components we are integrating the H fields over,
        # adjust weighting to account for pre-weighting of the source by `EPSILON_0`
        # and multiply by appropriate `MU_0` factor
        for H_perp in H_integration_components:
            contrib_H = MU_0 * H_perp / EPSILON_0
            vjps += contrib_H

        return vjps

    @staticmethod
    def _project_in_basis(
        field_components: dict[str, np.ndarray],
        basis_vector: np.ndarray,
    ) -> np.ndarray:
        """Project 3D field components onto a basis vector.

        Parameters
        ----------
        field_components : dict[str, np.ndarray]
            Dictionary with keys like "Ex", "Ey", "Ez" or "Dx", "Dy", "Dz" containing field values.
            Values have shape (N, F) where F is the number of frequencies.
        basis_vector : np.ndarray
            (N, 3) array of basis vectors, one per evaluation point.

        Returns
        -------
        np.ndarray
            Projected field values with shape (N, F).
        """
        prefix = next(iter(field_components.keys()))[0]
        field_matrix = np.stack([field_components[f"{prefix}{dim}"] for dim in "xyz"], axis=0)

        # always expect (3, N, F) shape, transpose to (N, 3, F)
        field_matrix = np.transpose(field_matrix, (1, 0, 2))
        return np.einsum("ij...,ij->i...", field_matrix, basis_vector)

    def project_der_map_to_axis(
        self, axis: xyz, field_type: str = "E"
    ) -> dict[str, ScalarFieldDataArray] | None:
        """Return a copy of the selected derivative map with only one axis kept.

        Parameters
        ----------
        axis:
            Axis to keep (``"x"``, ``"y"``, ``"z"``, case-insensitive).
        field_type:
            Map selector: ``"E"`` (``self.E_der_map``) or ``"D"`` (``self.D_der_map``).

        Returns
        -------
        dict[str, ScalarFieldDataArray] | None
            Copied map where non-selected components are replaced by zeros, or ``None``
            if the requested map is unavailable.
        """
        field_map = {"E": self.E_der_map, "D": self.D_der_map}.get(field_type)
        if field_map is None:
            raise ValueError("field type must be 'D' or 'E'.")

        axis = axis.lower()
        projected = dict(field_map)
        if not field_map:
            return projected
        for dim in "xyz":
            key = f"E{dim}"
            if key not in field_map:
                continue
            if dim != axis:
                projected[key] = _zeros_like(field_map[key])
            else:
                projected[key] = field_map[key]
        return projected

    def adaptive_vjp_spacing(
        self,
        wl_fraction: Optional[float] = None,
        min_allowed_spacing_fraction: Optional[float] = None,
    ) -> float:
        """Compute adaptive spacing for finite-difference gradient evaluation.

        Determines an appropriate spatial resolution based on the material
        properties and electromagnetic wavelength/skin depth.

        Parameters
        ----------
        wl_fraction : float, optional
            Fraction of wavelength/skin depth to use as spacing. Defaults to the configured
            ``autograd.default_wavelength_fraction`` when ``None``.
        min_allowed_spacing_fraction : float, optional
            Minimum allowed spacing fraction of free space wavelength used to
            prevent numerical issues. Defaults to ``config.adjoint.minimum_spacing_fraction``
            when not specified.

        Returns
        -------
        float
            Adaptive spacing value for gradient evaluation.
        """
        if wl_fraction is None or min_allowed_spacing_fraction is None:
            from tidy3d.config import config

            if wl_fraction is None:
                wl_fraction = config.adjoint.default_wavelength_fraction
            if min_allowed_spacing_fraction is None:
                min_allowed_spacing_fraction = config.adjoint.minimum_spacing_fraction

        def spacing_by_permittivity(eps_array):
            eps_real = np.asarray(eps_array.values, dtype=np.complex128).real

            dx_candidates = []
            max_frequency = np.max(self.frequencies)

            # wavelength-based sampling for dielectrics
            if np.any(eps_real > 0):
                eps_max = eps_real[eps_real > 0].max()
                lambda_min = self.wavelength_min / np.sqrt(eps_max)
                dx_candidates.append(wl_fraction * lambda_min)

            # skin depth sampling for metals
            if np.any(eps_real <= 0):
                omega = 2 * np.pi * max_frequency
                eps_neg = eps_real[eps_real <= 0]
                delta_min = C_0 / (omega * np.sqrt(np.abs(eps_neg).max()))
                dx_candidates.append(wl_fraction * delta_min)

            computed_spacing = min(dx_candidates)

            return computed_spacing

        eps_spacings = [
            spacing_by_permittivity(eps_array) for _, eps_array in self.eps_data.items()
        ]
        computed_spacing = np.min(eps_spacings)

        min_allowed_spacing = self.wavelength_min * min_allowed_spacing_fraction

        if computed_spacing < min_allowed_spacing:
            log.warning(
                f"Based on the material, the adaptive spacing for integrating the polyslab surface "
                f"would be {computed_spacing:.3e} μm. The spacing has been clipped to {min_allowed_spacing:.3e} μm "
                f"to prevent a performance degradation.",
                log_once=True,
            )

        return max(computed_spacing, min_allowed_spacing)

    @property
    def wavelength_min(self) -> float:
        return C_0 / np.max(self.frequencies)

    @property
    def wavelength_max(self) -> float:
        return C_0 / np.min(self.frequencies)


def integrate_within_bounds(arr: xr.DataArray, dims: list[str], bounds: Bound) -> xr.DataArray:
    """Integrate a data array within specified spatial bounds.

    Clips the integration domain to the specified bounds and performs
    numerical integration using the trapezoidal rule.

    Parameters
    ----------
    arr : xr.DataArray
        Data array to integrate.
    dims : list[str]
        Dimensions to integrate over (e.g., ['x', 'y', 'z']).
    bounds : Bound
        Integration bounds as [[xmin, ymin, zmin], [xmax, ymax, zmax]].

    Returns
    -------
    xr.DataArray
        Result of integration with specified dimensions removed.

    Notes
    -----
    - Coordinates outside bounds are clipped, effectively setting dL=0
    - Only integrates dimensions with more than one coordinate point
    - Uses xarray's integrate method (trapezoidal rule)
    """
    bounds = np.asarray(bounds).T
    all_coords = {}

    for dim, (bmin, bmax) in zip(dims, bounds):
        bmin = get_static(bmin)
        bmax = get_static(bmax)

        # clip coordinates to bounds (sets dL=0 outside bounds)
        coord_values = arr.coords[dim].data
        all_coords[dim] = np.clip(coord_values, bmin, bmax)

    _arr = arr.assign_coords(**all_coords)

    # only integrate dimensions with multiple points
    dims_integrate = [dim for dim in dims if len(_arr.coords[dim]) > 1]
    return _arr.integrate(coord=dims_integrate)


__all__ = [
    "DerivativeInfo",
    "integrate_within_bounds",
]



from __future__ import annotations

from abc import ABC

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.spice.analysis.dc import (
    IsothermalSteadyChargeDCAnalysis,
    SteadyChargeDCAnalysis,
)
from tidy3d.components.types import ArrayFloat1D
from tidy3d.constants import HERTZ
from tidy3d.constants import inf as td_inf


class AbstractSSACAnalysis(Tidy3dBaseModel, ABC):
    """
    Abstract base class for Small-Signal AC (SSAC) analysis parameters.

    Notes
    -----
    This class provides the common interface for SSAC analysis by adding the
    ``freqs`` field to any DC analysis class.

    Small-signal analysis is performed by linearizing the device equations
    around a DC operating point. The analysis computes the small-signal response
    at the specified frequencies.

    Examples
    --------
    >>> import tidy3d as td
    >>> freq_range = td.FreqRange.from_freq_interval(1e3, 1e6)
    >>> sweep_freqs = freq_range.sweep_decade(num_points_per_decade=10)
    >>> ssac_spec = td.SSACAnalysis(freqs=sweep_freqs)
    """

    freqs: ArrayFloat1D = pd.Field(
        ...,
        title="Small Signal AC Frequencies",
        description="List of frequencies for small signal AC analysis. "
        "At least one :class:`.SSACVoltageSource` must be present in the boundary conditions.",
        units=HERTZ,
    )

    @pd.validator("freqs")
    def validate_freqs(cls, val):
        if len(val) == 0:
            raise ValueError("'freqs' cannot be empty (size 0).")
        else:
            for freq in val:
                if freq == td_inf:
                    raise ValueError("'freqs' cannot contain infinite frequencies.")
                elif freq < 0:
                    raise ValueError("'freqs' cannot contain negative frequencies.")
        return val


class SSACAnalysis(SteadyChargeDCAnalysis, AbstractSSACAnalysis):
    """
    Configures Small-Signal AC (SSAC) analysis parameters for charge simulation.

    Examples
    --------
    >>> import tidy3d as td
    >>> freq_range = td.FreqRange.from_freq_interval(1e3, 1e6)
    >>> sweep_freqs = freq_range.sweep_decade(num_points_per_decade=10)
    >>> ssac_spec = td.SSACAnalysis(freqs=sweep_freqs)
    """


class IsothermalSSACAnalysis(IsothermalSteadyChargeDCAnalysis, AbstractSSACAnalysis):
    """
    Configures Isothermal Small-Signal AC (SSAC) analysis parameters for charge simulation.

    Notes
    -----
    This analysis class provides an interface for SSAC analysis.

    Examples
    --------
    >>> import tidy3d as td
    >>> freq_range = td.FreqRange.from_freq_interval(1e3, 1e6)
    >>> sweep_freqs = freq_range.sweep_decade(num_points_per_decade=10)
    >>> ssac_spec = td.IsothermalSSACAnalysis(freqs=sweep_freqs)
    """



"""Lumped port specialization with a rectangular geometry."""

from __future__ import annotations

from typing import Optional

import numpy as np
import pydantic.v1 as pd
from shapely import union_all
from shapely.geometry.base import BaseMultipartGeometry

from tidy3d.components.base import cached_property
from tidy3d.components.data.data_array import FreqDataArray
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.geometry.base import Box, Geometry
from tidy3d.components.geometry.utils import (
    SnapBehavior,
    SnapLocation,
    SnappingSpec,
    snap_box_to_grid,
)
from tidy3d.components.geometry.utils_2d import increment_float
from tidy3d.components.grid.grid import Grid, YeeGrid
from tidy3d.components.lumped_element import LinearLumpedElement, LumpedResistor, RLCNetwork
from tidy3d.components.medium import LossyMetalMedium, PECMedium
from tidy3d.components.microwave.path_integrals.integrals.current import AxisAlignedCurrentIntegral
from tidy3d.components.microwave.path_integrals.integrals.voltage import AxisAlignedVoltageIntegral
from tidy3d.components.monitor import FieldMonitor
from tidy3d.components.source.current import UniformCurrentSource
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.structure import Structure
from tidy3d.components.types import Axis, FreqArray, LumpDistType
from tidy3d.components.validators import assert_line_or_plane
from tidy3d.exceptions import SetupError, ValidationError

from .base_lumped import AbstractLumpedPort


class LumpedPort(AbstractLumpedPort, Box):
    """Class representing a single rectangular lumped port.

    Example
    -------
    >>> port1 = LumpedPort(center=(0, 0, 0),
    ...             size=(0, 1, 2),
    ...             voltage_axis=2,
    ...             name="port_1",
    ...             impedance=50
    ...         )

    See Also
    --------
    :class:`.LinearLumpedElement`
        The lumped element representing the load of the port.
    """

    voltage_axis: Axis = pd.Field(
        ...,
        title="Voltage Integration Axis",
        description="Specifies the axis along which the E-field line integral is performed when "
        "computing the port voltage. The integration axis must lie in the plane of the port.",
    )

    snap_perimeter_to_grid: bool = pd.Field(
        True,
        title="Snap Perimeter to Grid",
        description="When enabled, the perimeter of the port is snapped to the simulation grid, "
        "which improves accuracy when the number of grid cells is low within the element. A :class:`LumpedPort` "
        "is always snapped to the grid along its injection axis.",
    )

    dist_type: LumpDistType = pd.Field(
        "on",
        title="Distribute Type",
        description="Optional field that is passed directly to the :class:`.LinearLumpedElement` used to model the port's load. "
        "When set to ``on``, the network portion of the lumped port, including the source, is distributed"
        "across the entirety of the lumped element's bounding box. When set to ``off``, the network "
        "portion of the lumped port is restricted to one cell and PEC connections are used to "
        "connect the network cell to the edges of the lumped element. A third option exists "
        "``laterally_only``, where the network portion is only distributed along the lateral axis of "
        "the lumped port.",
    )

    _line_plane_validator = assert_line_or_plane()

    @cached_property
    def injection_axis(self):
        """Injection axis of the port."""
        return self.size.index(0.0)

    @pd.validator("voltage_axis", always=True)
    def _voltage_axis_in_plane(cls, val, values):
        """Ensure voltage integration axis is in the port's plane."""
        size = values.get("size")
        if val == size.index(0.0):
            raise ValidationError("'voltage_axis' must lie in the port's plane.")
        return val

    @cached_property
    def current_axis(self) -> Axis:
        """Integration axis for computing the port current via the magnetic field."""
        return 3 - self.injection_axis - self.voltage_axis

    def to_source(
        self, source_time: GaussianPulse, snap_center: Optional[float] = None, grid: Grid = None
    ) -> UniformCurrentSource:
        """Create a current source from the lumped port."""
        if grid:
            # This will included any snapping behavior the load undergoes
            load_box = self._to_load_box(grid=grid)
            center = load_box.center
            size = load_box.size
        else:
            # Discretized source amps are manually zeroed out later if they
            # fall on Yee grid locations outside the analytical source region.
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            size = self.size

        component = "xyz"[self.voltage_axis]
        return UniformCurrentSource(
            center=center,
            size=size,
            source_time=source_time,
            polarization=f"E{component}",
            name=self.name,
            interpolate=True,
            confine_to_bounds=True,
        )

    def to_load(self, snap_center: Optional[float] = None) -> LumpedResistor:
        """Create a load resistor from the lumped port."""
        # 2D materials are currently snapped to the grid, so snapping here is not needed.
        # It is done here so plots of the simulation will more accurately portray the setup
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center

        network = RLCNetwork(resistance=np.real(self.impedance))
        return LinearLumpedElement(
            center=center,
            size=self.size,
            num_grid_cells=self.num_grid_cells,
            network=network,
            name=f"{self.name}_resistor",
            voltage_axis=self.voltage_axis,
            snap_perimeter_to_grid=self.snap_perimeter_to_grid,
            dist_type=self.dist_type,
            enable_snapping_points=self.enable_snapping_points,
        )

    def to_voltage_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> FieldMonitor:
        """Field monitor to compute port voltage."""
        if grid:
            voltage_box = self._to_voltage_box(grid)
            center = voltage_box.center
            size = voltage_box.size
        else:
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            # Size of voltage monitor can essentially be 1D from ground to signal conductor
            size = list(self.size)
            size[self.injection_axis] = 0.0
            size[self.current_axis] = 0.0

        e_component = "xyz"[self.voltage_axis]
        # Create a voltage monitor
        return FieldMonitor(
            center=center,
            size=size,
            freqs=freqs,
            fields=[f"E{e_component}"],
            name=self._voltage_monitor_name,
            colocate=False,
        )

    def to_current_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> FieldMonitor:
        """Field monitor to compute port current."""
        if grid:
            current_box = self._to_current_box(grid)
            center = current_box.center
            size = current_box.size
        else:
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            # Size of current monitor needs to encompass the current carrying 2D sheet
            # Needs to have a nonzero thickness so a closed loop of gridpoints around
            # the 2D sheet can be formed
            dl = 2 * (
                increment_float(center[self.injection_axis], 1.0) - center[self.injection_axis]
            )
            size = list(self.size)
            size[self.injection_axis] = dl
            size[self.voltage_axis] = 0.0

        h_component = "xyz"[self.current_axis]
        h_cap_component = "xyz"[self.injection_axis]
        # Create a current monitor
        return FieldMonitor(
            center=center,
            size=size,
            freqs=freqs,
            fields=[f"H{h_component}", f"H{h_cap_component}"],
            name=self._current_monitor_name,
            colocate=False,
        )

    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""
        voltage_box = self._to_voltage_box(sim_data.simulation.grid)
        field_data = sim_data[self._voltage_monitor_name]
        voltage_integral = AxisAlignedVoltageIntegral(
            center=voltage_box.center,
            size=voltage_box.size,
            extrapolate_to_endpoints=True,
            snap_path_to_grid=True,
            sign="+",
        )
        voltage = voltage_integral.compute_voltage(field_data)
        # Return data array of voltage with coordinates of frequency
        return voltage

    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing through the port."""
        # Diagram of contour integral, dashed line indicates location of sheet resistance
        # and electric field used for voltage computation. Voltage axis is out-of-page.
        #
        #                                    current_axis = ->
        #                                    injection_axis = ^
        #
        #                  |                   h2_field ->             |
        #    h_cap_minus ^  -------------------------------------------  h_cap_plus ^
        #                  |                   h1_field ->             |

        field_data = sim_data[self._current_monitor_name]
        current_box = self._to_current_box(sim_data.simulation.grid)

        # H field is continuous at integral bounds, so extrapolation is turned off
        I_integral = AxisAlignedCurrentIntegral(
            center=current_box.center,
            size=current_box.size,
            sign="+",
            extrapolate_to_endpoints=True,
            snap_contour_to_grid=True,
        )
        return I_integral.compute_current(field_data)

    def _check_grid_size(self, yee_grid: YeeGrid) -> None:
        """Raises :class:`SetupError` if the grid is too coarse at port locations"""
        e_component = "xyz"[self.voltage_axis]
        e_yee_grid = yee_grid.grid_dict[f"E{e_component}"]
        coords = e_yee_grid.to_dict[e_component]
        min_bound = self.bounds[0][self.voltage_axis]
        max_bound = self.bounds[1][self.voltage_axis]
        coords_within_port = np.any(np.logical_and(coords > min_bound, coords < max_bound))
        if not coords_within_port:
            raise SetupError(
                f"Grid is too coarse along '{e_component}' direction for the lumped port "
                f"at location '{self.center}'. Either set the port's 'num_grid_cells' to "
                f"a nonzero integer or modify the 'GridSpec'."
            )

    def _to_load_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the exact location of the load,
        after it is snapped to the grid."""
        load = self.to_load()
        # This will included any snapping behavior the load undergoes
        load_box = load._create_box_for_network(grid=grid)
        return load_box

    def _to_voltage_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the location of the
        path integral for computing voltage."""
        load_box = self._to_load_box(grid=grid)
        size = list(load_box.size)
        size[self.current_axis] = 0
        size[self.injection_axis] = 0
        voltage_box = Box(center=load_box.center, size=size)
        return voltage_box

    def _to_current_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the location of the
        path integral for computing current."""
        load_box = self._to_load_box(grid=grid)
        size = list(load_box.size)
        size[self.voltage_axis] = 0
        current_box = Box(center=load_box.center, size=size)
        # Snap the current contour integral to the nearest magnetic field positions
        # that enclose the load box/sheet resistance
        snap_location = [SnapLocation.Center] * 3
        snap_behavior = [SnapBehavior.Expand] * 3
        snap_behavior[self.voltage_axis] = SnapBehavior.Off
        snap_spec = SnappingSpec(location=snap_location, behavior=snap_behavior)
        current_box = snap_box_to_grid(grid, current_box, snap_spec)
        return current_box

    @classmethod
    def from_structures(
        cls,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ground_terminal: Structure | None = None,
        signal_terminal: Structure | None = None,
        voltage_axis: Axis = None,
        lateral_coord: Optional[float] = None,
        port_width: Optional[float] = None,
        **kwargs,
    ) -> LumpedPort:
        """
        Auto-generate lumped port based on provided structures and plane coordinates.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        ground_terminal : Structure = None
            Structure representing ground terminal.
        signal_terminal : Structure = None
            Structure representing signal terminal.
        voltage_axis : Axis = None
            Direction of lumped port voltage.
        lateral_coord : Optional[float] = None
            Coordinate along lateral axis.
        port_width: float = None
            Lateral Width of lumped port.
        **kwargs:
            Other LumpedPort parameters.

        Returns
        -------
        LumpedPorts
            Lumped Port defined between ground and signal terminals.

        Notes
        -----
        - The lateral axis refers to the axis perpendicular to the voltage_axis within the port plane
        - lateral_coord has two purposes:
            (a) if ground/signal has multiple 2D shapes intersecting the port plane, the shape chosen to create the port is the one that intersects lateral_coord.
            (b) if port_width is also specified, lateral_coord is used as the center position of the port along that axis.
        - If port_width is not specified, port width is automatically chosen based on 2D bounds of the in-plane ground/signal shapes.
        """

        # Parse normal axis and coordinate
        normal_axis, normal_coord = Geometry.parse_xyz_kwargs(x=x, y=y, z=z)

        # make sure voltage axis was specified
        if voltage_axis is None:
            raise ValueError(
                "No voltage axis was provided. Please make sure a valid voltage axis is specified."
            )

        # make sure normal is orthogonal to voltage axis
        if normal_axis == voltage_axis:
            xyz_string = "xyz"
            raise ValueError(
                "Voltage axis must lie in the plane of the Lumped Port. \n"
                f"The provided voltage axis for the Lumped port is: {xyz_string[voltage_axis]}. \n"
                f"The inferred axis normal to the Lumped Port is: {xyz_string[normal_axis]}"
            )

        # get lateral axis (defined as axis within  Lumped Port's plane orthogonal to voltage axis)
        lateral_axis = 3 - normal_axis - voltage_axis

        if ground_terminal is None or signal_terminal is None:
            raise ValueError(
                "Signal and ground terminals are required. Please check that both are specified."
            )

        # form a list of allowed terminal materials
        ALLOWED_TERMINAL_MEDIA = (LossyMetalMedium, PECMedium)

        # make sure that terminals are made of PEC or lossy metal
        for name, terminal in (("ground", ground_terminal), ("signal", signal_terminal)):
            if not isinstance(terminal.medium, ALLOWED_TERMINAL_MEDIA):
                raise ValidationError(
                    f"Invalid {name} terminal medium: expected either 'PEC' or 'LossyMetalMedium'. "
                    f"Got: '{terminal.medium}'."
                )

        # 2D ground and signal shapes and axes
        grounds_2d = ground_terminal.geometry.intersections_plane(x=x, y=y, z=z)
        signals_2d = signal_terminal.geometry.intersections_plane(x=x, y=y, z=z)

        _, coords_2d = Geometry.pop_axis((0, 1, 2), normal_axis)
        lateral_axis_2d = coords_2d.index(lateral_axis)
        voltage_axis_2d = 1 - lateral_axis_2d

        # Check: structures intersects plane
        if len(grounds_2d) == 0:
            raise ValueError("Ground structure does not intersect Lumped Port plane.")
        if len(signals_2d) == 0:
            raise ValueError("Signal structure does not intersect Lumped Port plane.")

        # Get 2 element list with index 0 = ground shape and index 1 = signal shape
        shape_list = []

        # loop over polygons / contours of terminals cut by a Lumped Port plane
        for name, shapes in [("Ground", grounds_2d), ("Signal", signals_2d)]:
            sel_shape = None

            # If lateral_coord is specified, pick ground/signal shape based on intersection
            if lateral_coord is not None:
                # constract a line along voltage axis in the lumped port plane going throung the lateral coordinate
                if voltage_axis_2d == 0:
                    line = Geometry.make_shapely_box(-np.inf, lateral_coord, np.inf, lateral_coord)
                else:
                    line = Geometry.make_shapely_box(lateral_coord, -np.inf, lateral_coord, np.inf)

                # collect all geometries in shapes that intersect the line
                intersected_shapes = [s for s in shapes if s.intersects(line)]

                if not intersected_shapes:
                    raise ValidationError(
                        f"{name} terminal was not detected at the specified lateral coordinate {lateral_coord}. "
                        f"Please, make sure provided coordinate corresponds to a valid terminal location."
                    )
                # merge all loops of the terminal to one geometry
                sel_shape = union_all(intersected_shapes)
            else:
                # combine all shapely geometries and check if result is a polygon
                sel_shape = union_all(shapes)

            # Check Lumped Port plane intersects a terminal
            if sel_shape.is_empty:
                raise ValidationError(
                    f"{name} terminal is not intersecting the lumped port plane or lateral coordinate."
                    "Please make sure the lumped port plane and/or terminals are specified correctly."
                )
            if isinstance(sel_shape, BaseMultipartGeometry):
                coord_msg = (
                    f"at lateral coordinate {lateral_coord}" if lateral_coord is not None else ""
                )
                coord_hint = (
                    "Please define the port manually."
                    if lateral_coord is not None
                    else "Please specify a lateral coordinate to select the correct conductor."
                )
                raise ValidationError(
                    f"Automatic lumped port setup failed: more than one {name.lower()} terminal conductor "
                    f"intersects the lumped port plane {coord_msg}. {coord_hint}"
                )
            shape_list.append(sel_shape)

        # Get 2d shapes
        ground_2d, signal_2d = shape_list

        # ensure that signal and ground terminals do not intersect
        if ground_2d.intersects(signal_2d):
            raise ValidationError(
                "Ground intersects signal in the specified plane."
                "Please make sure that ground and signal terminals do not overlap."
            )

        # get terminals' bounds
        ground_bounds = np.array(ground_2d.bounds).reshape(2, 2).T
        signal_bounds = np.array(signal_2d.bounds).reshape(2, 2).T

        def intervals_overlap(a, b):
            """Return True if [a_min, a_max] and [b_min, b_max] overlap."""
            a_min, a_max = a
            b_min, b_max = b
            return not (a_max < b_min or a_min > b_max)

        # ensure that terminal bounding boxes don't overlap along voltage axis
        if intervals_overlap(ground_bounds[voltage_axis_2d], signal_bounds[voltage_axis_2d]):
            raise ValidationError(
                "Auto-generation of lumped port failed because ground and signal terminals have overlapping bounds along voltage axis. "
                "Please define lumped port manually."
            )
        # ensure that terminal bounding boxes overlap in lateral direction
        if not intervals_overlap(ground_bounds[lateral_axis_2d], signal_bounds[lateral_axis_2d]):
            raise ValidationError(
                "Auto-generation of lumped port failed because ground and signal terminals "
                "don't have overlapping bounds along lateral axis. "
                "Please define lumped port manually."
            )

        # reshape and combine bounds of the signal and ground cuts; rows correspond to axes
        bounds = np.hstack((ground_bounds, signal_bounds))

        # for each axis extract port bounds
        new_bounds = np.sort(bounds, axis=1)[:, 1:3]
        vmin, vmax = new_bounds[voltage_axis_2d, :]
        lmin, lmax = new_bounds[lateral_axis_2d, :]

        if port_width is not None:
            if lateral_coord is not None:
                # If port_width and lateral_coord are specified, set based on those instead
                lmin_new, lmax_new = (
                    lateral_coord - port_width / 2,
                    lateral_coord + port_width / 2,
                )
            else:
                # If only port_width is specified, set based on bounds
                lcenter = (lmin + lmax) / 2
                lmin_new, lmax_new = (lcenter - port_width / 2, lcenter + port_width / 2)
                # make sure that port_width does not exceed width of terminal overlap along lateral axis.

            if lmin_new < lmin or lmax_new > lmax:
                raise ValueError(
                    "Specified port region extends beyond the overlap between the signal and ground terminals. "
                    "Please apply appropriate 'port_width' and 'lateral_coord', or set them to 'None' to allow for automatic assignment."
                )
            # update port limits along lateral axis
            lmin, lmax = lmin_new, lmax_new

        # WARNING: This currently does not check if edges of lumped port touches ground/signal.
        #          This can happen if ground/signal does not overlap along lateral_axis
        #          It is up to the user to provide sensible geometry choices for ground/signal to avoid this

        # allocate memory for lumped port bounds
        rmin = np.zeros(3)
        rmax = np.zeros(3)

        # construct lumped port bounds
        rmin[voltage_axis], rmax[voltage_axis] = (vmin, vmax)
        rmin[lateral_axis], rmax[lateral_axis] = (lmin, lmax)
        rmin[normal_axis], rmax[normal_axis] = (normal_coord, normal_coord)

        kwargs["voltage_axis"] = voltage_axis
        lumped_port = LumpedPort.from_bounds(rmin=tuple(rmin), rmax=tuple(rmax), **kwargs)

        return lumped_port



"""Stub for webapi"""

from __future__ import annotations

from datetime import datetime
from os import PathLike
from typing import Callable, Optional

import pydantic.v1 as pd
from pydantic.v1 import BaseModel

from tidy3d import log
from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.data.monitor_data import ModeSolverData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.eme.simulation import EMESimulation
from tidy3d.components.microwave.data.monitor_data import MicrowaveModeSolverData
from tidy3d.components.mode.data.sim_data import ModeSimulationData
from tidy3d.components.mode.simulation import ModeSimulation
from tidy3d.components.simulation import Simulation
from tidy3d.components.tcad.mesher import VolumeMesher
from tidy3d.components.tcad.simulation.heat import HeatSimulation
from tidy3d.components.tcad.simulation.heat_charge import HeatChargeSimulation
from tidy3d.components.types.workflow import WorkflowDataType, WorkflowType
from tidy3d.plugins.mode.mode_solver import ModeSolver
from tidy3d.plugins.smatrix.component_modelers.modal import (
    ModalComponentModeler,
)
from tidy3d.plugins.smatrix.component_modelers.terminal import (
    TerminalComponentModeler,
)
from tidy3d.plugins.smatrix.data.modal import (
    ModalComponentModelerData,
)
from tidy3d.plugins.smatrix.data.terminal import (
    TerminalComponentModelerData,
)
from tidy3d.web.core.stub import TaskStub, TaskStubData
from tidy3d.web.core.types import TaskType

TYPE_MAP: dict[type, TaskType] = {
    Simulation: TaskType.FDTD,
    ModeSolver: TaskType.MODE_SOLVER,
    HeatSimulation: TaskType.HEAT,
    HeatChargeSimulation: TaskType.HEAT_CHARGE,
    EMESimulation: TaskType.EME,
    ModeSimulation: TaskType.MODE,
    VolumeMesher: TaskType.VOLUME_MESH,
    ModalComponentModeler: TaskType.MODAL_CM,
    TerminalComponentModeler: TaskType.TERMINAL_CM,
}


def task_type_name_of(simulation: WorkflowType) -> str:
    for cls, ttype in TYPE_MAP.items():
        if isinstance(simulation, cls):
            return ttype.name
    raise TypeError(f"Could not find task type for: {type(simulation).__name__}")


class Tidy3dStub(BaseModel, TaskStub):
    simulation: WorkflowType = pd.Field(discriminator="type")

    @classmethod
    def from_file(cls, file_path: PathLike) -> WorkflowType:
        """Loads a ``WorkflowType`` instance from .yaml, .json, or .hdf5 file.

        Parameters
        ----------
        file_path : PathLike
            Full path to the .yaml or .json or .hdf5 file to load the
            ``WorkflowType`` from.

        Returns
        -------
        WorkflowType
            An instance of the component class calling ``load``.
        """
        return Tidy3dBaseModel.from_file(file_path)

    def to_file(
        self,
        file_path: PathLike,
    ) -> None:
        """Exports ``WorkflowType`` instance to .yaml, .json, or .hdf5 file

        Parameters
        ----------
        file_path : PathLike
            Full path to the .yaml or .json or .hdf5 file to save the ``WorkflowType`` to.
        """
        self.simulation.to_file(file_path)

    def to_hdf5_gz(self, fname: PathLike) -> None:
        """Exports ``WorkflowType`` instance to .hdf5.gz file.

        Parameters
        ----------
        fname : PathLike
            Full path to the .hdf5.gz file to save the ``WorkflowType`` to.
        """

        self.simulation.to_hdf5_gz(fname)

    def get_type(self) -> str:
        """Get simulation instance type.

        Returns
        -------
        :class:`TaskType`
            An instance Type of the component class calling ``load``.
        """
        return task_type_name_of(self.simulation)

    def validate_pre_upload(self, source_required: bool) -> None:
        """Perform some pre-checks on instances of component"""
        if isinstance(self.simulation, Simulation):
            self.simulation.validate_pre_upload(source_required)
        else:
            self.simulation.validate_pre_upload()

    def get_default_task_name(self) -> str:
        """
        Generate a default task name based on the simulation type and
        the current date and time.

        The name is composed of the simulation type and a human-readable timestamp in the format ``YYYY-MM-DD_HH-MM-SS``

        Example
        -------
        >>> stub.get_default_task_name() # doctest: +SKIP
        'fdtd_2025-09-16_14-30-55'

        Returns
        -------
        str
            Default task name, e.g. ``"fdtd_2025-09-16_14-30-55"``.
        """
        sim_type = self.get_type().lower()
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        return f"{sim_type}_{timestamp}"


class Tidy3dStubData(BaseModel, TaskStubData):
    """"""

    data: WorkflowDataType

    @classmethod
    def from_file(
        cls, file_path: PathLike, lazy: bool = False, on_load: Optional[Callable] = None
    ) -> WorkflowDataType:
        """Loads a ``WorkflowDataType`` instance from .yaml, .json, or .hdf5 file.

        Parameters
        ----------
        file_path : PathLike
            Full path to the .yaml or .json or .hdf5 file to load the ``WorkflowDataType`` instance
            from.
        lazy : bool = False
            Whether to load the actual data (``lazy=False``) or return a proxy that loads
            the data when accessed (``lazy=True``).
        on_load : Callable | None = None
            Callback function executed once the model is fully materialized.
            Only used if ``lazy=True``. The callback is invoked with the loaded
            instance as its sole argument, enabling post-processing such as
            validation, logging, or warnings checks.

        Returns
        -------
        ``WorkflowDataType`` instance
            An instance of the component class calling ``load``.
        """
        return Tidy3dBaseModel.from_file(file_path, lazy=lazy, on_load=on_load)

    def to_file(self, file_path: PathLike) -> None:
        """Exports ``WorkflowDataType`` instance
        to .yaml, .json, or .hdf5 file

        Parameters
        ----------
        file_path : PathLike
            Full path to the .yaml or .json or .hdf5 file to save the ``WorkflowDataType`` instance to.
        """
        self.data.to_file(file_path)

    @classmethod
    def postprocess(cls, file_path: PathLike, lazy: bool = True) -> WorkflowDataType:
        """Load .yaml, .json, or .hdf5 file to
        ``WorkflowDataType`` instance.

        Parameters
        ----------
        file_path : PathLike
            Full path to the .yaml or .json or .hdf5 file to save the ``WorkflowDataType`` instance to.
        lazy : bool = False
            Whether to load the actual data (``lazy=False``) or return a proxy that loads
            the data when accessed (``lazy=True``).

        Returns
        -------
        ``WorkflowDataType`` instance
            An instance of the component class calling ``load``.
        """
        workflow_data = Tidy3dBaseModel.from_file(
            file_path, lazy=lazy, on_load=cls._check_convergence_and_warnings
        )
        return workflow_data

    @staticmethod
    def _check_convergence_and_warnings(stub_data: WorkflowDataType) -> None:
        """Check convergence, divergence, and warnings in the solver log and emit log messages."""
        check_log_msg = (
            "For more information, check 'SimulationData.log' or use 'web.download_log(task_id)'."
        )
        warned_about_warnings = False

        if isinstance(stub_data, SimulationData):
            final_decay_value = stub_data.final_decay_value
            shutoff_value = stub_data.simulation.shutoff
            if stub_data.diverged:
                log.warning("The simulation has diverged! " + check_log_msg)
                warned_about_warnings = True
            elif (shutoff_value != 0) and (final_decay_value > shutoff_value):
                log.warning(
                    f"Simulation final field decay value of {final_decay_value} is greater than "
                    f"the simulation shutoff threshold of {shutoff_value}. Consider running the "
                    "simulation again with a larger 'run_time' duration for more accurate results."
                )

        if (
            not isinstance(
                stub_data,
                (
                    ModeSolverData,
                    MicrowaveModeSolverData,
                    ModeSimulationData,
                    TerminalComponentModelerData,
                    ModalComponentModelerData,
                ),
            )
            and "WARNING" in stub_data.log
            and not warned_about_warnings
        ):
            log.warning("Warning messages were found in the solver log. " + check_log_msg)



# container for everything defining the inverse design

from __future__ import annotations

import abc
import typing
from typing import Any

import autograd.numpy as anp
import numpy as np
import pydantic.v1 as pd

import tidy3d as td
from tidy3d.components.autograd import get_static
from tidy3d.exceptions import ValidationError
from tidy3d.plugins.expressions.metrics import Metric, generate_validation_data
from tidy3d.plugins.expressions.types import ExpressionType

from .base import InvdesBaseModel
from .region import DesignRegionType
from .validators import check_pixel_size

PostProcessFnType = typing.Callable[[td.SimulationData], float]


class AbstractInverseDesign(InvdesBaseModel, abc.ABC):
    """Container for an inverse design problem."""

    design_region: DesignRegionType = pd.Field(
        ...,
        title="Design Region",
        description="Region within which we will optimize the simulation.",
    )

    task_name: str = pd.Field(
        ...,
        title="Task Name",
        description="Task name to use in the objective function when running the ``JaxSimulation``.",
    )

    verbose: bool = pd.Field(
        False,
        title="Task Verbosity",
        description="If ``True``, will print the regular output from ``web`` functions.",
    )

    metric: typing.Optional[ExpressionType] = pd.Field(
        None,
        title="Objective Metric",
        description="Serializable expression defining the objective function.",
    )

    def make_objective_fn(
        self, post_process_fn: typing.Optional[typing.Callable] = None, maximize: bool = True
    ) -> typing.Callable[[anp.ndarray], tuple[float, dict]]:
        """Construct the objective function for this InverseDesign object."""

        if (post_process_fn is None) and (self.metric is None):
            raise ValueError("Either 'post_process_fn' or 'metric' must be provided.")

        if (post_process_fn is not None) and (self.metric is not None):
            raise ValueError("Provide only one of 'post_process_fn' or 'metric', not both.")

        direction_multiplier = 1 if maximize else -1

        def objective_fn(params: anp.ndarray, aux_data: typing.Optional[dict] = None) -> float:
            """Full objective function."""
            data = self.to_simulation_data(params=params)

            if self.metric is None:
                post_process_val = post_process_fn(data)
            elif isinstance(data, td.SimulationData):
                post_process_val = self.metric.evaluate(data)
            elif getattr(data, "type", None) == "BatchData":
                raise NotImplementedError("Metrics currently do not support 'BatchData'")
            else:
                raise ValueError(f"Invalid data type: {type(data)}")

            penalty_value = self.design_region.penalty_value(params)
            objective_fn_val = direction_multiplier * post_process_val - penalty_value

            # Store auxiliary data if provided
            if aux_data is not None:
                aux_data["penalty"] = get_static(penalty_value)
                aux_data["post_process_val"] = get_static(post_process_val)
                aux_data["objective_fn_val"] = get_static(objective_fn_val) * direction_multiplier
                if isinstance(data, td.SimulationData):
                    aux_data["sim_data"] = data.to_static()
                else:
                    aux_data["sim_data"] = {k: v.to_static() for k, v in data.items()}
                aux_data["params"] = params

            return objective_fn_val

        return objective_fn

    @property
    def initial_simulation(self) -> td.Simulation:
        """Return a simulation with the initial design region parameters."""
        initial_params = self.design_region.initial_parameters
        return self.to_simulation(initial_params)

    def run(self, simulation, **kwargs: Any) -> td.SimulationData:
        """Run a single tidy3d simulation."""
        from tidy3d.web import run

        kwargs.setdefault("verbose", self.verbose)
        kwargs.setdefault("task_name", self.task_name)
        return run(simulation, **kwargs)

    def run_async(self, simulations, **kwargs: Any) -> web.BatchData:  # noqa: F821
        """Run a batch of tidy3d simulations."""
        from tidy3d.web import run_async

        kwargs.setdefault("verbose", self.verbose)
        return run_async(simulations, **kwargs)


class InverseDesign(AbstractInverseDesign):
    """Container for an inverse design problem."""

    simulation: td.Simulation = pd.Field(
        ...,
        title="Base Simulation",
        description="Simulation without the design regions or monitors used in the objective fn.",
    )

    output_monitor_names: tuple[str, ...] = pd.Field(
        None,
        title="Output Monitor Names",
        description="Optional names of monitors whose data the differentiable output depends on."
        "If this field is left ``None``, the plugin will try to add all compatible monitors "
        "automatically. While this will work, there may be warnings if certain monitor types are "
        "not fully supported, for example ``FieldMonitor`` instances with ``.colocate != False``.",
    )

    _check_sim_pixel_size = check_pixel_size("simulation")

    @pd.root_validator(pre=False)
    def _validate_model(cls, values: dict) -> dict:
        cls._validate_metric(values)
        return values

    @staticmethod
    def _validate_metric(values: dict) -> dict:
        metric_expr = values.get("metric")
        if not metric_expr:
            return values
        simulation = values.get("simulation")
        for metric in metric_expr.filter(Metric):
            InverseDesign._validate_metric_monitor_name(metric, simulation)
            InverseDesign._validate_metric_mode_index(metric, simulation)
            InverseDesign._validate_metric_f(metric, simulation)
        InverseDesign._validate_metric_data(metric_expr, simulation)
        return values

    @staticmethod
    def _validate_metric_monitor_name(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the monitor name of the metric exists in the simulation."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if monitor is None:
            raise ValidationError(
                f"Monitor named '{metric.monitor_name}' associated with the metric not found in the simulation monitors."
            )

    @staticmethod
    def _validate_metric_mode_index(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the mode index of the metric is within the bounds of the monitor's ``ModeSpec.num_modes``."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if metric.mode_index >= monitor.mode_spec.num_modes:
            raise ValidationError(
                f"Mode index '{metric.mode_index}' for metric associated with monitor "
                f"'{metric.monitor_name}' is out of bounds. "
                f"Maximum allowed mode index is '{monitor.mode_spec.num_modes - 1}'."
            )

    @staticmethod
    def _validate_metric_f(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the frequencies of the metric are present in the monitor."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if metric.f is not None:
            metric_f_list = [metric.f] if isinstance(metric.f, float) else metric.f
            if len(metric_f_list) != 1:
                raise ValidationError("Only a single frequency is supported for the metric.")
            for freq in metric_f_list:
                if not any(np.isclose(freq, monitor.freqs, atol=1.0)):
                    raise ValidationError(
                        f"Frequency '{freq}' for metric associated with monitor "
                        f"'{metric.monitor_name}' not found in monitor frequencies."
                    )
        else:
            if len(monitor.freqs) != 1:
                raise ValidationError(
                    f"Monitor '{metric.monitor_name}' must contain only a single frequency when metric.f is None."
                )

    @staticmethod
    def _validate_metric_data(expr: ExpressionType, simulation: td.Simulation) -> None:
        """Validate that expression can be evaluated and returns a real scalar."""
        data = generate_validation_data(expr)
        try:
            result = expr(data)
        except Exception as e:
            raise ValidationError(f"Failed to evaluate the metric expression: {e!s}") from e
        if len(np.ravel(result)) > 1:
            raise ValidationError(
                f"The expression must return a scalar value or an array of length 1 (got {result})."
            )
        if not np.all(np.isreal(result)):
            raise ValidationError(
                f"The expression must return a real (not complex) value (got {result})."
            )

    def is_output_monitor(self, monitor: td.Monitor) -> bool:
        """Whether a monitor is added to the ``JaxSimulation`` as an ``output_monitor``."""

        output_mnt_types = td.components.simulation.OutputMonitorTypes

        if self.output_monitor_names is None:
            return any(isinstance(monitor, mnt_type) for mnt_type in output_mnt_types)

        return monitor.name in self.output_monitor_names

    def separate_output_monitors(self, monitors: tuple[td.Monitor]) -> dict:
        """Separate monitors into output_monitors and regular monitors."""

        monitor_fields = {"monitors": [], "output_monitors": []}

        for monitor in monitors:
            key = "output_monitors" if self.is_output_monitor(monitor) else "monitors"
            monitor_fields[key].append(monitor)

        return monitor_fields

    def to_simulation(self, params: anp.ndarray) -> td.Simulation:
        """Convert the ``InverseDesign`` to a corresponding ``td.Simulation`` with traced fields."""

        # construct the design region to a regular structure
        design_region_structure = self.design_region.to_structure(params)

        # construct mesh override structures and a new grid spec, if applicable
        grid_spec = self.simulation.grid_spec
        mesh_override_structure = self.design_region.mesh_override_structure
        if mesh_override_structure:
            override_structures = list(self.simulation.grid_spec.override_structures)
            override_structures += [mesh_override_structure]
            grid_spec = grid_spec.updated_copy(override_structures=override_structures)

        return self.simulation.updated_copy(
            structures=[*list(self.simulation.structures), design_region_structure],
            grid_spec=grid_spec,
        )

    def to_simulation_data(self, params: anp.ndarray, **kwargs: Any) -> td.SimulationData:
        """Convert the ``InverseDesign`` to a ``td.Simulation`` and run it."""
        simulation = self.to_simulation(params=params)
        return self.run(simulation, **kwargs)


class InverseDesignMulti(AbstractInverseDesign):
    """``InverseDesign`` with multiple simulations and corresponding postprocess functions."""

    simulations: tuple[td.Simulation, ...] = pd.Field(
        ...,
        title="Base Simulations",
        description="Set of simulation without the design regions or monitors used in the objective fn.",
    )

    output_monitor_names: tuple[typing.Union[tuple[str, ...], None], ...] = pd.Field(
        None,
        title="Output Monitor Names",
        description="Optional names of monitors whose data the differentiable output depends on."
        "If this field is left ``None``, the plugin will try to add all compatible monitors "
        "automatically. While this will work, there may be warnings if certain monitor types are "
        "not fully supported, for example ``FieldMonitor`` instances with ``.colocate != False``.",
    )

    _check_sim_pixel_size = check_pixel_size("simulations")

    @pd.root_validator()
    def _check_lengths(cls, values):
        """Check the lengths of all of the multi fields."""

        keys = ("simulations", "post_process_fns", "output_monitor_names", "override_structure_dl")
        multi_dict = {key: values.get(key) for key in keys}
        sizes = {key: len(val) for key, val in multi_dict.items() if val is not None}

        if len(set(sizes.values())) != 1:
            raise ValueError(
                f"'MultiInverseDesign' requires that the fields {keys} must either "
                "have the same length or be left ``None``, if optional. Given fields with "
                "corresponding sizes of '{sizes}'."
            )

        return values

    @property
    def task_names(self) -> list[str]:
        """Task names associated with each of the simulations."""
        return [f"{self.task_name}_{i}" for i in range(len(self.simulations))]

    @property
    def designs(self) -> list[InverseDesign]:
        """List of individual ``InverseDesign`` objects corresponding to this instance."""

        designs_list = []
        for i, (task_name, sim) in enumerate(zip(self.task_names, self.simulations)):
            des_i = InverseDesign(
                design_region=self.design_region,
                simulation=sim,
                verbose=self.verbose,
                task_name=task_name,
            )
            if self.output_monitor_names is not None:
                des_i = des_i.updated_copy(output_monitor_names=self.output_monitor_names[i])

            designs_list.append(des_i)

        return designs_list

    def to_simulation(self, params: anp.ndarray) -> dict[str, td.Simulation]:
        """Convert the ``InverseDesign`` to a corresponding dict of ``td.Simulation``s."""
        simulation_list = [design.to_simulation(params) for design in self.designs]
        return dict(zip(self.task_names, simulation_list))

    def to_simulation_data(self, params: anp.ndarray, **kwargs: Any) -> web.BatchData:  # noqa: F821
        """Convert the ``InverseDesignMulti`` to a set of ``td.Simulation``s and run async."""
        simulations = self.to_simulation(params)
        return self.run_async(simulations, **kwargs)


InverseDesignType = typing.Union[InverseDesign, InverseDesignMulti]



# utility functions for autograd web API
from __future__ import annotations

import typing

import numpy as np

import tidy3d as td

""" E and D field gradient map calculation helpers. """


def get_derivative_maps(
    fld_fwd: td.FieldData,
    eps_fwd: td.PermittivityData,
    fld_adj: td.FieldData,
    eps_adj: td.PermittivityData,
) -> dict[str, td.FieldData]:
    """Get electric and displacement field derivative maps."""
    der_map_E = derivative_map_E(fld_fwd=fld_fwd, fld_adj=fld_adj)
    der_map_D = derivative_map_D(fld_fwd=fld_fwd, eps_fwd=eps_fwd, fld_adj=fld_adj, eps_adj=eps_adj)

    make_H_der_map = np.all([f"H{dim}" in fld_fwd.field_components for dim in "xyz"])
    der_map_H = None
    if make_H_der_map:
        der_map_H = derivative_map_H(fld_fwd=fld_fwd, fld_adj=fld_adj)

    return {"E": der_map_E, "D": der_map_D, "H": der_map_H}


def derivative_map_E(fld_fwd: td.FieldData, fld_adj: td.FieldData) -> td.FieldData:
    """Get td.FieldData where the Ex, Ey, Ez components store the gradients w.r.t. these."""
    return multiply_field_data(fld_fwd, fld_adj, fld_key="E")


def derivative_map_H(fld_fwd: td.FieldData, fld_adj: td.FieldData) -> td.FieldData:
    """Get td.FieldData where the Hx, Hy, Hz components store the gradients w.r.t. these."""
    return multiply_field_data(fld_fwd, fld_adj, fld_key="H")


def derivative_map_D(
    fld_fwd: td.FieldData,
    eps_fwd: td.PermittivityData,
    fld_adj: td.FieldData,
    eps_adj: td.PermittivityData,
) -> td.FieldData:
    """Get td.FieldData where the Ex, Ey, Ez components store the gradients w.r.t. D fields."""
    fwd_D = E_to_D(fld_data=fld_fwd, eps_data=eps_fwd)
    adj_D = E_to_D(fld_data=fld_adj, eps_data=eps_adj)

    return multiply_field_data(fwd_D, adj_D, fld_key="E")


def E_to_D(fld_data: td.FieldData, eps_data: td.PermittivityData) -> td.FieldData:
    """Convert electric field to displacement field."""

    return multiply_field_data(fld_data, eps_data, fld_key="E")


def multiply_field_data(
    fld_1: td.FieldData, fld_2: typing.Union[td.FieldData, td.PermittivityData], fld_key: str
) -> td.FieldData:
    """Elementwise multiply two field data objects, writes data into ``fld_1`` copy."""

    def get_field_key(dim: str, fld_data: typing.Union[td.FieldData, td.PermittivityData]) -> str:
        """Get the key corresponding to the scalar field along this dimension."""
        return f"{fld_key}{dim}" if isinstance(fld_data, td.FieldData) else f"eps_{dim}{dim}"

    field_components = {}
    for dim in "xyz":
        key_1 = get_field_key(dim=dim, fld_data=fld_1)
        key_2 = get_field_key(dim=dim, fld_data=fld_2)
        cmp_1 = fld_1.field_components[key_1]
        cmp_2 = fld_2.field_components[key_2]
        mult = cmp_1 * cmp_2
        field_components[key_1] = mult
    return fld_1.updated_copy(**field_components)



"""
Our DC sources ultimately need to follow this standard form if we want to enable full electrical integration.

```
11.3.2 .DC: DC Transfer Function

General form:

    .dc srcnam vstart vstop vincr [src2 start2 stop2 incr2]

Examples:

    .dc VIN 0.25 5.0 0.25
    .dc VDS 0 10 .5 VGS 0 5 1
    .dc VCE 0 10 .25 IB 0 10u 1u
    .dc RLoad 1k 2k 100
    .dc TEMP -15 75 5
```

"""

from __future__ import annotations

from typing import Literal, Optional

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import ArrayFloat1D
from tidy3d.constants import AMP, VOLT
from tidy3d.constants import inf as td_inf
from tidy3d.log import log


class DCVoltageSource(Tidy3dBaseModel):
    """
    DC voltage source in volts.

    Notes
    -----

        This voltage refers to potential above the equivalent simulation ground. Currently, electrical ports
        are not defined.

    Examples
    --------
    >>> import tidy3d as td
    >>> voltages = [-0.5, 0, 1, 2, 3, 4]
    >>> voltage_source = td.DCVoltageSource(voltage=voltages)
    """

    name: Optional[str] = pd.Field(
        None,
        title="Name",
        description="Unique name for the DC voltage source",
        min_length=1,
    )

    voltage: ArrayFloat1D = pd.Field(
        ...,
        title="Voltage",
        description="DC voltage usually used as source in :class:`VoltageBC` boundary conditions.",
        units=VOLT,
    )

    # TODO: This should have always been in the field above but was introduced wrongly as a
    # standalone field. Keeping for compatibility, remove in 3.0.
    units: Literal[VOLT] = VOLT

    @pd.validator("voltage")
    def check_voltage(cls, val):
        for v in val:
            if v == td_inf:
                raise ValueError(f"Voltages must be finite. Currently  voltage={val}.")
        return val

    @staticmethod
    def _count_unique_with_tolerance(arr, rtol=1e-9, atol=1e-12):
        """Count unique values treating values within tolerance as duplicates.

        Uses sorted comparison to group values that are practically equal
        due to floating-point representation differences (e.g., single vs double precision).
        """
        if len(arr) == 0:
            return 0
        sorted_arr = np.sort(arr)
        # Count values that are "different enough" from their predecessor
        unique_count = 1
        for i in range(1, len(sorted_arr)):
            if not np.isclose(sorted_arr[i], sorted_arr[i - 1], rtol=rtol, atol=atol):
                unique_count += 1
        return unique_count

    @pd.validator("voltage")
    def check_repeated_voltage(cls, val):
        """Warn if repeated voltage values are present, treating 0 and -0 as the same value.

        Uses tolerance-based comparison to handle floating-point representation
        differences (e.g., values from single vs double precision sources).
        """
        # Normalize all zero values (both 0.0 and -0.0) to 0.0 so they are treated as duplicates
        normalized = np.where(np.isclose(val, 0, atol=1e-10), 0.0, val)
        unique_count = cls._count_unique_with_tolerance(normalized)
        if unique_count < len(val):
            log.warning(
                "Duplicate voltage values detected in 'voltage' array. "
                f"Found {len(val)} values but only {unique_count} are unique. "
                "Note: values within floating-point tolerance are considered duplicates."
            )
        return val


class GroundVoltage(Tidy3dBaseModel):
    """
    Ground voltage source (0V reference).


    Notes
    -----
    This source explicitly sets the ground reference (0V) for the simulation.
    It is equivalent to :class:`DCVoltageSource(voltage=0)` but more explicit about
    establishing the ground reference.

    If no :class:`GroundVoltage` is specified, the smallest voltage among all
    sources will be considered as the ground reference. Note that the boundary
    conditions defined using a voltage array will be ignored during this
    process and cannot be used as a default ground.

    Example
    -------
    >>> import tidy3d as td
    >>> ground_source = td.GroundVoltage()
    >>> voltage_bc = td.VoltageBC(source=ground_source)
    """


class DCCurrentSource(Tidy3dBaseModel):
    """
    DC current source in amperes.

    Example
    -------
    >>> import tidy3d as td
    >>> current_source = td.DCCurrentSource(current=0.4)
    """

    name: Optional[str] = pd.Field(
        None,
        title="Name",
        description="Unique name for the DC current source",
        min_length=1,
    )

    current: pd.FiniteFloat = pd.Field(
        title="Current",
        description="DC current usually used as source in :class:`CurrentBC` boundary conditions.",
        units=AMP,
    )

    # TODO: This should have always been in the field above but was introduced wrongly as a
    # standalone field. Keeping for compatibility, remove in 3.0.
    units: Literal[AMP] = AMP



"""EME dataset"""

from __future__ import annotations

from typing import Optional

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.data.data_array import (
    EMECoefficientDataArray,
    EMEFluxDataArray,
    EMEInterfaceSMatrixDataArray,
    EMEModeIndexDataArray,
    EMEScalarFieldDataArray,
    EMEScalarModeFieldDataArray,
    EMESMatrixDataArray,
)
from tidy3d.components.data.dataset import Dataset, ElectromagneticFieldDataset
from tidy3d.exceptions import ValidationError


class EMESMatrixDataset(Dataset):
    """Dataset storing S matrix."""

    S11: EMESMatrixDataArray = pd.Field(
        ...,
        title="S11 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 1.",
    )
    S12: EMESMatrixDataArray = pd.Field(
        ...,
        title="S12 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 2.",
    )
    S21: EMESMatrixDataArray = pd.Field(
        ...,
        title="S21 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 1.",
    )
    S22: EMESMatrixDataArray = pd.Field(
        ...,
        title="S22 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 2.",
    )


class EMEInterfaceSMatrixDataset(Dataset):
    """Dataset storing S matrices associated with EME cell interfaces."""

    S11: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="S11 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 1.",
    )
    S12: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="S12 matrix",
        description="S matrix relating output modes at port 1 to input modes at port 2.",
    )
    S21: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="S21 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 1.",
    )
    S22: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="S22 matrix",
        description="S matrix relating output modes at port 2 to input modes at port 2.",
    )


class EMEOverlapDataset(Dataset):
    """Dataset storing overlaps between EME modes.

    Notes
    -----
        ``Oij`` is the unconjugated overlap computed using the E field of cell ``i``
        and the H field of cell ``j``.

        For consistency with ``Sij``, ``mode_index_out`` refers to the mode index
        in cell ``i``, and ``mode_index_in`` refers to the mode index in cell ``j``.
    """

    O11: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="O11 matrix",
        description="Overlap integral between E field and H field in the same cell.",
    )
    O12: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="O12 matrix",
        description="Overlap integral between E field on side 1 and H field on side 2.",
    )
    O21: EMEInterfaceSMatrixDataArray = pd.Field(
        ...,
        title="O21 matrix",
        description="Overlap integral between E field on side 2 and H field on side 1.",
    )


class EMECoefficientDataset(Dataset):
    """Dataset storing various coefficients related to the EME simulation.

    Notes
    -----
        These coefficients can be used for debugging or optimization.

        The ``A`` and ``B`` fields store the expansion coefficients for the modes in a cell.
        These are defined at the cell centers.

        The ``n_complex`` and ``flux`` fields respectively store the complex-valued effective
        propagation index and the power flux associated with the modes used in the
        EME calculation.

        The ``interface_Sij`` fields store the S matrices associated with the interfaces
        between EME cells.
    """

    A: Optional[EMECoefficientDataArray] = pd.Field(
        None,
        title="A coefficient",
        description="Coefficient for forward mode in this cell.",
    )

    B: Optional[EMECoefficientDataArray] = pd.Field(
        None,
        title="B coefficient",
        description="Coefficient for backward mode in this cell.",
    )

    n_complex: Optional[EMEModeIndexDataArray] = pd.Field(
        None,
        title="Propagation Index",
        description="Complex-valued effective propagation indices associated with the EME modes.",
    )

    flux: Optional[EMEFluxDataArray] = pd.Field(
        None,
        title="Flux",
        description="Power flux of the EME modes.",
    )

    interface_smatrices: Optional[EMEInterfaceSMatrixDataset] = pd.Field(
        None,
        title="Interface S Matrices",
        description="S matrices associated with the interfaces between EME cells.",
    )

    overlaps: Optional[EMEOverlapDataset] = pd.Field(
        None, title="Overlaps", description="Overlaps between EME modes."
    )

    @cached_property
    def normalized_copy(self) -> EMECoefficientDataset:
        """Return a copy of the ``EMECoefficientDataset`` where
        the ``A`` and ``B`` coefficients as well as the ``interface_smatrices``
        are normalized by flux."""
        if self.flux is None:
            raise ValidationError(
                "The 'flux' field of the 'EMECoefficientDataset' is 'None', "
                "so normalization cannot be performed."
            )
        fields = {"A": self.A, "B": self.B}
        flux_out = self.flux.rename(mode_index="mode_index_out")
        for key, field in fields.items():
            if field is not None:
                fields[key] = field * np.sqrt(flux_out)
        if self.interface_smatrices is not None:
            num_cells = len(self.flux.eme_cell_index)
            flux1 = self.flux.isel(eme_cell_index=np.arange(num_cells - 1))
            flux2 = self.flux.isel(eme_cell_index=np.arange(1, num_cells))
            flux2 = flux2.assign_coords(eme_cell_index=np.arange(num_cells - 1))
            interface_S12 = self.interface_smatrices.S12
            flux_out = flux1.rename(mode_index="mode_index_out")
            flux_in = flux2.rename(mode_index="mode_index_in")
            interface_S12 = interface_S12 * np.sqrt(flux_out / flux_in)
            interface_S21 = self.interface_smatrices.S21
            flux_out = flux2.rename(mode_index="mode_index_out")
            flux_in = flux1.rename(mode_index="mode_index_in")
            interface_S21 = interface_S21 * np.sqrt(flux_out / flux_in)

            fields["interface_smatrices"] = self.interface_smatrices.updated_copy(
                S12=interface_S12, S21=interface_S21
            )
        # for safety to prevent normalizing twice
        fields["flux"] = None
        return self.updated_copy(**fields)


class EMEFieldDataset(ElectromagneticFieldDataset):
    """Dataset storing scalar components of E and H fields as a function of freq, mode_index, and port_index."""

    Ex: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field of the mode.",
    )
    Ey: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field of the mode.",
    )
    Ez: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field of the mode.",
    )
    Hx: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field of the mode.",
    )
    Hy: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field of the mode.",
    )
    Hz: EMEScalarFieldDataArray = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field of the mode.",
    )


class EMEModeSolverDataset(ElectromagneticFieldDataset):
    """Dataset storing EME modes as a function of freq, mode_index, and cell_index."""

    n_complex: EMEModeIndexDataArray = pd.Field(
        ...,
        title="Propagation Index",
        description="Complex-valued effective propagation constants associated with the mode.",
    )

    Ex: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field of the mode.",
    )
    Ey: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field of the mode.",
    )
    Ez: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field of the mode.",
    )
    Hx: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field of the mode.",
    )
    Hy: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field of the mode.",
    )
    Hz: EMEScalarModeFieldDataArray = pd.Field(
        ...,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field of the mode.",
    )



"""Base classes for performing path integrals with fields on the Yee grid"""

from __future__ import annotations

from typing import Literal, Union

import numpy as np
import xarray as xr

from tidy3d.components.data.data_array import (
    IntegralResultType,
    ScalarFieldDataArray,
    ScalarFieldTimeDataArray,
    ScalarModeFieldDataArray,
    _make_base_result_data_array,
)
from tidy3d.components.data.monitor_data import FieldData, FieldTimeData, ModeData, ModeSolverData
from tidy3d.components.microwave.path_integrals.specs.base import (
    AxisAlignedPathIntegralSpec,
    Custom2DPathIntegralSpec,
)
from tidy3d.constants import fp_eps
from tidy3d.exceptions import DataError

IntegrableMonitorDataType = Union[FieldData, FieldTimeData, ModeData, ModeSolverData]
EMScalarFieldType = Union[ScalarFieldDataArray, ScalarFieldTimeDataArray, ScalarModeFieldDataArray]
FieldParameter = Literal["E", "H"]


class AxisAlignedPathIntegral(AxisAlignedPathIntegralSpec):
    """Class for defining the simplest type of path integral, which is aligned with Cartesian axes.

    Example
    -------
    >>> path = AxisAlignedPathIntegral(
    ...     center=(0, 0, 1),
    ...     size=(0, 0, 2),
    ...     extrapolate_to_endpoints=True,
    ...     snap_path_to_grid=False,
    ... )
    """

    def compute_integral(self, scalar_field: EMScalarFieldType) -> IntegralResultType:
        """Computes the defined integral given the input ``scalar_field``."""

        if not scalar_field.does_cover(self.bounds, fp_eps, np.finfo(np.float32).smallest_normal):
            raise DataError("Scalar field does not cover the integration domain.")
        coord = "xyz"[self.main_axis]

        scalar_field = self._get_field_along_path(scalar_field)
        # Get the boundaries
        min_bound = self.bounds[0][self.main_axis]
        max_bound = self.bounds[1][self.main_axis]

        if self.extrapolate_to_endpoints:
            # Remove field outside the boundaries
            scalar_field = scalar_field.sel({coord: slice(min_bound, max_bound)})
            # Ignore values on the boundary (sel is inclusive)
            scalar_field = scalar_field.drop_sel({coord: (min_bound, max_bound)}, errors="ignore")
            coordinates = scalar_field.coords[coord].values
        else:
            coordinates = scalar_field.coords[coord].sel({coord: slice(min_bound, max_bound)})

        # Integration is along the original coordinates plus ensure that
        # endpoints corresponding to the precise bounds of the port are included
        coords_interp = np.array([min_bound])
        coords_interp = np.concatenate((coords_interp, coordinates))
        coords_interp = np.concatenate((coords_interp, [max_bound]))
        coords_interp = {coord: coords_interp}

        # Use extrapolation for the 2 additional endpoints, unless there is only a single sample point
        method = "linear"
        if len(coordinates) == 1 and self.extrapolate_to_endpoints:
            method = "nearest"
        scalar_field = scalar_field.interp(
            coords_interp, method=method, kwargs={"fill_value": "extrapolate"}
        )
        result = scalar_field.integrate(coord=coord)
        return _make_base_result_data_array(result)

    def _get_field_along_path(self, scalar_field: EMScalarFieldType) -> EMScalarFieldType:
        """Returns a selection of the input ``scalar_field`` ready for integration."""
        (axis1, axis2) = self.remaining_axes
        (coord1, coord2) = self.remaining_dims

        if self.snap_path_to_grid:
            # Coordinates that are not integrated
            remaining_coords = {
                coord1: self.center[axis1],
                coord2: self.center[axis2],
            }
            # Select field nearest to center of integration line
            scalar_field = scalar_field.sel(
                remaining_coords,
                method="nearest",
                drop=False,
            )
        else:
            # Try to interpolate unless there is only a single coordinate
            coord1dict = {coord1: self.center[axis1]}
            if scalar_field.sizes[coord1] == 1:
                scalar_field = scalar_field.sel(coord1dict, method="nearest")
            else:
                scalar_field = scalar_field.interp(
                    coord1dict, method="linear", kwargs={"bounds_error": True}
                )
            coord2dict = {coord2: self.center[axis2]}
            if scalar_field.sizes[coord2] == 1:
                scalar_field = scalar_field.sel(coord2dict, method="nearest")
            else:
                scalar_field = scalar_field.interp(
                    coord2dict, method="linear", kwargs={"bounds_error": True}
                )
        # Remove unneeded coordinates
        scalar_field = scalar_field.reset_coords(drop=True)
        return scalar_field

    @staticmethod
    def _check_monitor_data_supported(em_field: IntegrableMonitorDataType) -> None:
        """Helper for validating that monitor data is supported."""
        if not isinstance(em_field, (FieldData, FieldTimeData, ModeData, ModeSolverData)):
            supported_types = list(IntegrableMonitorDataType.__args__)
            raise DataError(
                f"'em_field' type {type(em_field)} not supported. Supported types are "
                f"{supported_types}"
            )


class Custom2DPathIntegral(Custom2DPathIntegralSpec):
    """Class for defining a custom path integral defined as a curve on an axis-aligned plane.

    Notes
    -----

    Given a set of vertices :math:`\\vec{r}_i`, this class approximates path integrals over
    vector fields of the form :math:`\\int{\\vec{F} \\cdot \\vec{dl}}`
    as :math:`\\sum_i{\\vec{F}(\\vec{r}_i) \\cdot \\vec{dl}_i}`,
    where the differential length :math:`\\vec{dl}` is approximated using central differences
    :math:`\\vec{dl}_i = \\frac{\\vec{r}_{i+1} - \\vec{r}_{i-1}}{2}`.
    If the path is not closed, forward and backward differences are used at the endpoints.

    Example
    -------
    >>> import numpy as np
    >>> vertices = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
    >>> path = Custom2DPathIntegral(
    ...     axis=2,
    ...     position=0.5,
    ...     vertices=vertices,
    ... )
    """

    def compute_integral(
        self, field: FieldParameter, em_field: IntegrableMonitorDataType
    ) -> IntegralResultType:
        """Computes the path integral defined by ``vertices`` given the input ``em_field``.

        Parameters
        ----------
        field : :class:`.FieldParameter`
            Can take the value of ``"E"`` or ``"H"``. Determines whether to perform the integral
            over electric or magnetic field.
        em_field : :class:`.IntegrableMonitorDataType`
            The electromagnetic field data that will be used for integrating.

        Returns
        -------
        :class:`.IntegralResultType`
            Result of integral over remaining dimensions (frequency, time, mode indices).
        """

        (dim1, dim2, dim3) = self.local_dims

        h_field_name = f"{field}{dim1}"
        v_field_name = f"{field}{dim2}"

        # Validate that fields are present
        em_field._check_fields_stored([h_field_name, v_field_name])

        # Select fields lying on the plane
        plane_indexer = {dim3: self.position}
        field1 = em_field.field_components[h_field_name].sel(plane_indexer, method="nearest")
        field2 = em_field.field_components[v_field_name].sel(plane_indexer, method="nearest")

        # Although for users we use the convention that an axis is simply `popped`
        # internally we prefer a right-handed coordinate system where dimensions
        # keep a proper order. The only change is to swap 'x' and 'z' when the
        # normal axis is along  `y`
        # Dim 's' represents the parameterization of the line
        # 't' is likely used for time
        if self.main_axis == 1:
            x_path = xr.DataArray(self.vertices[:, 1], dims="s")
            y_path = xr.DataArray(self.vertices[:, 0], dims="s")
        else:
            x_path = xr.DataArray(self.vertices[:, 0], dims="s")
            y_path = xr.DataArray(self.vertices[:, 1], dims="s")

        path_indexer = {dim1: x_path, dim2: y_path}
        field1_interp = field1.interp(path_indexer, method="linear")
        field2_interp = field2.interp(path_indexer, method="linear")

        # Determine the differential length elements along the path
        dl_x = Custom2DPathIntegralSpec._compute_dl_component(x_path, self.is_closed_contour)
        dl_y = Custom2DPathIntegralSpec._compute_dl_component(y_path, self.is_closed_contour)
        dl_x = xr.DataArray(dl_x, dims="s")
        dl_y = xr.DataArray(dl_y, dims="s")

        # Compute the dot product between differential length element and vector field
        integrand = field1_interp * dl_x + field2_interp * dl_y
        # Integrate along the path
        result = integrand.integrate(coord="s")
        result = result.reset_coords(drop=True)
        return _make_base_result_data_array(result)
